# INTEL CONFIDENTIAL
# Copyright 2014-2017 Intel Corporation All Rights Reserved.
#
# The source code contained or described herein and all
# documents related to the source code ("Material") are
# owned by Intel Corporation or its suppliers or licensors.
# Title to the Material remains with Intel Corporation or
# its suppliers and licensors. The Material may contain trade
# secrets and proprietary and confidential information of Intel
# Corporation and its suppliers and licensors, and is protected
# by worldwide copyright and trade secret laws and treaty
# provisions. No part of the Material may be used, copied,
# reproduced, modified, published, uploaded, posted, transmitted,
# distributed, or disclosed in any way without Intel's prior express
# written permission.
#
# No license under any patent, copyright, trade secret or other
# intellectual property right is granted to or conferred upon you
# by disclosure or delivery of the Materials, either expressly, by
# implication, inducement, estoppel or otherwise. Any license under
# such intellectual property rights must be express and approved by
# Intel in writing.
#
# Unless otherwise agreed by Intel in writing, you may not remove or
# alter this notice or any other notice embedded in Materials by Intel
# or Intel's suppliers or licensors in any way.
from lib.diags import *
from lib.hw import HwError, Devices
from ctypes import *

def read_cmd_complete(obj, text, line, start_index, end_index):
    return register_name_complete(obj.hw, text, line, start_index, end_index,
                                  obj.hw.get_register_names())

def read_cmd_help():
    return {"synopsis": "read <register spec> | (<addr> <length>)",
            "description":
            "Read a HW register or a HW memory address and display " + \
            "content. In the case of reading a register, only the " + \
            "requested register will be read and displayed. In the case of " + \
            "reading HW memory, the specified number of bytes will be read, " + \
            "provided that <cnt> is a multiple of 8.",
            "arguments":
            ["The command accepts both symbolic register specifications (see " + \
             "\"Addressing\" in the help) and HW memory addresses.", "",
             [("<register spec>", "CSR specification as described in the " + \
               "\"Addressing\" help section. Using this form will read the " + \
               "value of one or more registers depending on the values given " + \
               "in the argument and display the complete value, as well as the " + \
               "decoded bitfields."),
              ("<addr>", "When using the memory address form of the command, " + \
               "this is the address at which the read will start."),
              ("<length>", "When using the memory address form, the command " + \
               "will read this many bytes of memory starting at address <addr>. " + \
               "<length> must be a multiple of 8bytes. The command will not " + \
               "perform any decoding of the read memory and will display it in" + \
               "raw form.")]
             ],
            "returns":
            "On successfull read, the command will display the value of the " + \
            "read register (if <reg name> form used) or the content of the " + \
            "memory starting at address <addr>.\n" + \
            "If an error occurs during the read operation(s), an appropriate " + \
            "error message will be displayed.",
            "see also":
            "Addressing, write, decode"}

def read_cmd(obj, line, register_output):
    if not line:
        obj.error("Invalid read command")
        return False
    words = line.split()
    if is_hex(words[0]):
        # we need an address and a length
        if len(words) <= 2:
            addr = int(words[0], 16)
            if len(words) == 1: size = 8
            elif is_number(words[1]): size = str2num(words[1])
            else:
                obj.error("Invalid argument for <size>")
                return False
            if not size:
                obj.error("Invalid size value")
                return False
            try: result = obj.hw.read(c_ulonglong(addr), c_ulong(size))
            except HwError, e:
                obj.hw_error(e)
                return False

            prefix = "%#0x: "
            output = []
            if not isinstance(result, (list, tuple)): result = [result]
            for idx in range(0, len(result), 4):
                values = tuple(result[idx:idx+4])
                args = (addr,) + values
                format = prefix + "%#018x " * len(values)
                output.append(format % args)
                addr += 4 * 8
            obj.show_cmd_output(output)
        else:
            obj.error("Invalid argument list")
    else:
        reginfo = []
        if not validate_register_spec(words[0], reginfo, obj.hw):
            obj.error(reginfo[0])
            return False

        for csr in reginfo:
            reg, cs, ce, rs, re, field = csr
            for idx in range(cs, ce+1):
                for elem in range(rs, re+1):
                    try: result = obj.hw.read_register(reg, idx, elem)
                    except HwError, e:
                        obj.hw_error(str(e).strip())
                        break

                    addr = obj.hw.addressof(reg, idx, elem)
                    if obj.config.show_register_bitfields:
                        decoded = obj.hw.decode_register(reg, result)
                        if field: show_fields = [getattr(decoded, field)]
                        else: show_fields = decoded.fields
                    else: show_fields = None
                    freg = obj.format_register(reg, idx, elem, addr, result,
                                               show_fields)
                    if freg: register_output.append(freg + "\n")
    return False

def write_cmd_help():
    return {"synopsis": "write <register spec> | (<addr> <length>) <value>",
            "description":
            "Write a value to a HW register or HW memory. In the case where " + \
            "the command is to be writing to device memory, the same value " + \
            "will be written in every 8byte interval starting at <addr> and " + \
            "ending at <addr> + (<len> - 1).\n" + \
            "When the first argument is a HW register field, only that field " + \
            "of the register will be set. The rest of the bits in the register " + \
            "will be left unchanged (the operation is a read-modify-write).",
            "arguments":
            ["The command accepts both register specification (see \"Addressing\" " + \
             "in the help) and HW memory addresses.", "",
             [("<register spec>", "This form will write the value of one or more " + \
               "registers depending on the values given in the argument."),
              ("<value>", "The value to be written. This argument is required " + \
               "regardless of which form is being used."),
              ("<addr>", "When using the memory address form of the command, " + \
               "this is the address at which to start writing."),
              ("<length>", "When using the memory address form of the command, " + \
               "the value will be written in 8byte segments until total bytes " + \
               "written reaches <length> bytes.")],
             ],
            "returns":
            "On successful writing of all memory locations (registers or memory " + \
            "addresses), the command will not produce any output. If an error " + \
            "occured during one of the writes, the command will display an " + \
            "appropriate error message."}

def write_cmd(obj, line, output):
    if not line:
        obj.error("Invalid write command")
        return False
    cmdline = line.split()
    if is_hex(cmdline[0]):
        # Write to memory address
        if len(cmdline) == 2:
            baseaddr, length, value = int(cmdline[0], 16), "8", cmdline[1]
        elif len(cmdline) == 3:
            baseaddr, length, value = int(cmdline[0], 16), cmdline[1], cmdline[2]
        else:
            obj.error("Invalid number of arguments for memory write")
            return False

        if is_number(length):
            if is_hex(length): length = int(length, 16)
            else: length = int(length)
        else:
            obj.error("Only numeric values accepted for 'length' argument");
            return False
        if is_number(value):
            if is_hex(value): value = int(value, 16)
            else: value = int(value)
        else:
            obj.error("Only numeric values accepted for 'value' argument");
            return False
        try:
            obj.hw.write(c_ulonglong(baseaddr), c_ulonglong(value), c_ulong(length))
        except HwError, e:
            obj.hw_error(str(e).strip())
        return False
    else:
        if len(cmdline) != 2:
            obj.error("Invalid argument list. See 'help write'.")
            return False

        if is_number(cmdline[1]):
            if is_hex(cmdline[1]): value = c_ulonglong(int(cmdline[1], 16))
            else: value = c_ulonglong(int(cmdline[1]))
        else:
            obj.error("Invalid write value '%s'" % cmdline[1])
            return False

        reginfo = []
        if not validate_register_spec(cmdline[0], reginfo, obj.hw):
            obj.error(reginfo[0])
            return False

        for csr in reginfo:
            reg, cs, ce, rs, re, field = csr
            for idx in range(cs, ce+1):
                for elem in range(rs, re+1):
                    try:
                        if field: ret = obj.hw.write_register_field(reg, idx, elem,
                                                                    field, value)
                        else: ret = obj.hw.write_register(reg, idx, elem, value)
                    except HwError, e:
                        obj.hw_error(e)
                        break
    return False

def show_cmd_complete(obj, text, line, start_index, end_index):
    opts = ["globals", "global", "registers", "registerinfo", "all"]
    words = line.split()
    if len(words) == 2: return find_matches(text, opts)
    elif len(words) > 2:
        if text:
            if text != words[1]:
                if words[1] == "global":
                    return find_matches(text, obj.hw.get_global_names())
                elif words[1] == "register":
                    return find_matches(text, obj.hw.get_register_names())
        else:
            if words[1] == "global": return obj.hw.get_global_names()
            elif words[1] == "register": return obj.hw.get_register_names()
    return opts

def show_cmd_help():
    return {"synopsis":
            "show globals | global <definition> | registers | " + \
            "registerinfo | all",
            "description":
            "Show the global definitions and registers supported by the " + \
            "HW. This effectively describes the entire HW register " + \
            "space."}

def show_cmd(obj, line, output):
    if not line: line = "all"
    cmd = line.split()
    if not len(cmd): stype = "all"
    else: stype = cmd[0]

    if stype in ("globals", "all"):
        globdefs = obj.hw.get_globals()
        output += ["GLOBAL DEFINITIONS", HEADER]
        for glob in globdefs:
            output.append((glob[0], glob[1].value))

    if stype == "registers":
        if output: output.append(EMPTY_LINE)
        output += ["DEFINED REGISTERS", HEADER]
        output += obj.text2columns(obj.hw.get_register_names() +
                                   obj.hw.get_config_register_names())
    elif stype in ("registerinfo", "all"):
        if output: output.append(EMPTY_LINE)
        output += ["FULL REGISTER INFORMATION", HEADER]
        for reg in obj.hw.get_register_names() + \
                obj.hw.get_config_register_names():
            info_cmd(obj, reg, output)

    if stype == "global":
        if len(cmd) != 2:
            obj.error("Invalid command. See 'help show'.")
            return
        globdefs = obj.hw.get_globals()
        for glob in globdefs:
            if glob[0] != cmd[1]: continue
            output.append((glob[0], glob[1].value))
    return False

def info_cmd_complete(obj, text, line, start_index, end_index):
    return register_name_complete(obj.hw, text, line, start_index, end_index)

def info_cmd_help():
    return {"synopsis": "info <register spec>",
            "description":
            "Describe a HW register. This will show the address of the " + \
            "register and all of the bitfields in that register with " + \
            "their mask, shift, and smask values."}

def info_cmd(obj, line, output):
    cmdline = line.split()
    if not cmdline:
        obj.error("Missing command arguments")
        return False
    regname = cmdline[0]
    reginfo = []

    if is_hex(regname):
        reg = obj.hw.get_register_by_addr(int(regname, 16))
        info = reg.describe(0)
        output.append(obj.format_register_def(info[0], info[1], 0, info[4]) + "\n")
    else:
        if not validate_register_spec(regname, reginfo, obj.hw,
                                      obj.hw.get_register_names() +
                                      obj.hw.get_config_register_names()):
            obj.error(reginfo[0])
            return False

        for csr in reginfo:
            regname, cs, ce, rs, re, field = csr
            reg = obj.hw.get_register_by_name(regname)

            if not reg:
                obj.hw_error("Unknown register %s" % regname)
                return False

            # We only show the first register in the array. The rest are all the same
            info = reg.describe(0)

            output.append(obj.format_register_def(info[0], info[1], 0, info[4]) + "\n")
    return False

def unit_cmd(obj, line, output):
    words = line.split()
    if not line:
        obj.error("Missing unit number")
        return False

    if len(words) > 2:
        obj.error("Invalid options for command")
        return False

    devs = Devices()
    if words[0] == "list":
        output.append("Available units:")
        for device in devs:
            output.append("\tUnit %d: Type %s" % device)
        return False

    unit = str2num(words[0])
    if unit == None:
        obj.error("Invalid device number")
        return False

    newdev = devs.createDevice(unit)
    if not newdev:
        obj.error("Could not determine device type for device %d" % unit)
        return False

    obj.hw.close()
    del obj.hw
    obj.hw = newdev

    try:
        if not obj.hw.open():
            obj.error("Failed to open HFI%d device %d" % (obj.hw.get_hw(), unit))
        else:
            obj.show_cmd_output(["Using unit %d HFI%d" % (unit, obj.hw.get_hw())])
    except HwError, e: obj.hw_error(e)

    if obj.hw.load_hw_data():
        obj.error("Failed to load HW data for unit %d" % unit)
    return False

def unit_cmd_help():
    return {"synopsis": "unit <unit num> | list",
            "description":
            "Change the HFI unit on which the commands operate. <unit num> " + \
            "should correspond to a present unit, starting from 0. " + \
            "\"list\" shows all available units."}

__commands__ = [("read", "r", read_cmd, read_cmd_complete, read_cmd_help),
                ("write", "w", write_cmd, read_cmd_complete, write_cmd_help),
                ("show", show_cmd, show_cmd_complete, show_cmd_help),
                ("info", info_cmd, info_cmd_complete, info_cmd_help),
                ("unit", unit_cmd, None, unit_cmd_help)]
