# INTEL CONFIDENTIAL
# Copyright 2014-2016 Intel Corporation All Rights Reserved.
#
# The source code contained or described herein and all
# documents related to the source code ("Material") are
# owned by Intel Corporation or its suppliers or licensors.
# Title to the Material remains with Intel Corporation or
# its suppliers and licensors. The Material may contain trade
# secrets and proprietary and confidential information of Intel
# Corporation and its suppliers and licensors, and is protected
# by worldwide copyright and trade secret laws and treaty
# provisions. No part of the Material may be used, copied,
# reproduced, modified, published, uploaded, posted, transmitted,
# distributed, or disclosed in any way without Intel's prior express
# written permission.
#
# No license under any patent, copyright, trade secret or other
# intellectual property right is granted to or conferred upon you
# by disclosure or delivery of the Materials, either expressly, by
# implication, inducement, estoppel or otherwise. Any license under
# such intellectual property rights must be express and approved by
# Intel in writing.
#
# Unless otherwise agreed by Intel in writing, you may not remove or
# alter this notice or any other notice embedded in Materials by Intel
# or Intel's suppliers or licensors in any way.
import ctypes
import struct
from lib.diags import *
from lib.hw import HwError, HwStates, SupportedHWRevs

STATE_COMMANDS = ["save", "load", "capture", "diff", "switch", "show"]
DC8051_STATE_NAME = "DC8051"
DC8051_DATA_MEM_SIZE = 0x1000
DC8051_FW_VER_ADDR = 0x3a

JTAG_WARNING = ["WARNING: Capturing HW states while JTAG HW is attached",
                "WARNING: and DFT clocks are programmed will cause CSR",
                "WARNING: reads to hang. Please, ensure that DFT clocks",
                "WARNING: are not programmed and JTAG HW is not attached",
                "WARNING: before performing HW state capture."]

def describe_help():
    return {"synopsis": "status",
            "description":
            "Show status information for the HW device.\n\n"
            }

def dc8051_mem_file(hw):
    return "/sys/kernel/debug/hfi1/hfi1_%d/dc8051_memory" % hw.unit

def describe_hfi(obj, line, output):
    dcfirm = 0
    try:
        num_ctxt = obj.hw.read_register("SendContexts")
        num_sdma = obj.hw.read_register("SendDmaEngines")
        num_rcv_ctxt = obj.hw.read_register("RcvContexts")
        num_max_sctxt = obj.hw.get_global("TXE_NUM_CONTEXTS")
        num_max_rctxt = obj.hw.get_global("RXE_NUM_CONTEXTS")
        num_max_sdma = obj.hw.get_global("TXE_NUM_SDMA_ENGINES")
        info = obj.hw.hw_revision()

        try:
            with open(dc8051_mem_file(obj.hw), "rb", 0) as mem_fd:
                mem_fd.seek(DC8051_FW_VER_ADDR, 0)
                buf = mem_fd.read(2)
                dcfirm = struct.unpack("H", buf)[0]
        except IOError:
            print("Unable to read the 8051 firmware version")

    except HwError, e:
        obj.hw_error(str(e))
        return False

    output += ["HFI Information", SPACER,
               ("Implementation", info["impl_str"]),
               ("Chip Revision", "%u.%u (%s)" % \
                (info["chip_rev_major"], info["chip_rev_minor"],
                 SupportedHWRevs.byval(info["chip_rev_minor"]))),
               ("Implementation Revision", "0x%x" % info["impl_rev"]),
               ("DC8051 Firmware Version", "%d.%d" % \
                (((dcfirm >> 8) & 0xff), (dcfirm & 0xff))),
               ("Maximum Supported Send Contexts", num_max_sctxt),
               ("Number of Send Contexts", num_ctxt),
               ("Maximum Supported Rcv Contexts", num_max_rctxt),
               ("Number of Rcv Contexts", num_rcv_ctxt),
               ("Maximum Supported SDMA Engines", num_max_sdma),
               ("Number of SDMA Engines", num_sdma)]

    register_set =[]
    rset = [__r for __r in find_matches("*Status", obj.hw.get_register_names(), True, 0) \
            if "dbg" not in __r.lower()]
    for reg in rset:
        reg = obj.hw.get_register(reg)
        if reg and not reg.is_debug:
            register_set.append(reg)

    if not register_set: return False

    output += ["", "Chip Status", SPACER]

    globcsrs = []
    ctxtcsrs = []
    sdmacsrs = []

    for reg in register_set:
        if reg.count == 1: globcsrs.append(reg)
        elif reg.count == (num_max_sctxt - 1): ctxtcsrs.append(reg)
        elif reg.count == (num_max_sdma - 1): sdmacsrs.append(reg)

    # Chip Level CSRs
    for reg in globcsrs:
        value = obj.hw.read_register(reg)
        output += [(reg.name, obj.format_numeric_value(value))]

    extra_data = []
    for idx in range(num_max_sctxt):
        hdone = False
        for reg in ctxtcsrs:
            value = obj.hw.read_register(reg, idx)
            if value:
                if not hdone:
                    extra_data.append("Ctxt %d:" % idx)
                    hdone = True
                extra_data.append(("\t%s" % reg.name,
                                   obj.format_numeric_value(value)))
    output += extra_data

    extra_data = []
    for idx in range(num_max_sdma):
        hdone = False
        for reg in sdmacsrs:
            value = obj.hw.read_register(reg, idx)
            if value:
                if not hdone:
                    extra_data.append("SDE %d:" % idx)
                    hdone = True
                extra_data.append(("\t%s" % reg.name,
                                   obj.format_numeric_value(value)))
    output += extra_data
    return False

def help_state():
    return {"synopsis":
            "state <cmd> [<cmd args>]",
            "description":
            "The state command is used to capture, save, and or use " + \
            "a complete hardware state of the entire chip. The state " + \
            "captured will include the entire memory mapped chip area " + \
            "which includes all memory mapped CSRs.\n\nThe hardware " + \
            "state is kept internally and can be saved to a file or " + \
            "loaded from a file (see Arguments).\n\nThe internal memory " + \
            "object supports two seperate loaded memory states in " + \
            "addition to a single captured state. Each state capture will " + \
            "overwrite any previously captured state. Loading of " + \
            "additional state from files will overwrite previously " + \
            "loaded states in a round-robin fashion once both internal " + \
            "slots have been filled.",
            "arguments":
            ["The 'state' command support a number of sub-commands listed " + \
             "below. The arguments to each of the sub-commands vary.", "",
             "capture",
             "This form of the command will capture the " + \
             "hardware state to internal memory. This will not overwrite " + \
             "any loaded states. Successive captures will, however, overwrite " + \
             "each other.",
             "",
             "load <filename>",
             "Load a saved state into one of the internal memory slots. " + \
             "Once loaded, it can be made the active state and examined.", "",
             [("<filename>", "The file containing the saved state to load")],
             "",
             "save [<state>] <filename>",
             "This form of the command will save an internally captured state to " + \
             "a file. This file can later be loaded and used to examine the " + \
             "state of the hardware.", "",
             [("<state>", "Optional parameter specifying which state to be " + \
               "saved. When the argument is omitted, a newly captured state " + \
               "will be saved."),
              ("<filename>", "The file which the state will be saved into")],
             "",
             "diff <state A> <state B>",
             "Examine both hardware states and show any differences in CSR " + \
             "values.",
             [("<state A>", "The first of the two states to compare"),
              ("<state B>", "The second of the two states to compare")],
             "",
             "Both arguments can be any of the following: ",
             [("saved1", "State in the first internal slot"),
              ("saved2", "State in the second internal slot"),
              ("hw", "The current hardware state (this option effectively " + \
               "captures the current state prior to the comparison)")],
             "",
             "switch <state>",
             "Switch the active state to <state> (values possible for " + \
             "<stateN> are the same as with the 'diff' argument). When " + \
             "the state is switched to one of the internal states, the " + \
             "'read' command will get CSR values from that saved state.",
             "",
             "show",
             "This command argument will display information about all " + \
             "hardware states currently in memory (loaded and captured)."],
            "see also":
            "read"
            }

def complete_state(obj, text, line, start_index, end_index):
    words = line.split()
    if len(words) == 1: return STATE_COMMANDS
    elif len(words) == 2 and text: return find_matches(text, STATE_COMMANDS)
    elif len(words) >= 2:
        states = [str(x) for x in HwStates]
        if words[1] in ("load", "save"):
            # completion of state names is only needed when we are saving a
            # state
            if text and words[1] == "save":
                state_matches = find_matches(text, states + [DC8051_STATE_NAME])
                if state_matches: return state_matches
            return path_name_complete(text, line, start_index, end_index)
        elif words[1] == "switch":
            return find_matches(text, states)
        elif words[1] == "diff":
            if len(words) < 4 or (len(words) == 4 and text):
                return find_matches(text, states)

def state_save(obj, line, output):
    words = line.split()

    if not len(words):
        obj.error("Incomplete command")
        return False

    if words[0] not in STATE_COMMANDS:
        obj.error("Unkown argument '%s' for 'state' command" % words[0])
        return False

    if words[0] in ("save", "load"):
        if words[0] == "load" and len(words) != 2:
            obj.error("Missing filename")
            return False
        if words[0] == "save":
            if len(words) == 3:
                state, filename = words[1], words[2]
                if state.upper() not in HwStates:
                    # Handle DC8051 special case
                    if state.upper() == DC8051_STATE_NAME:
                        if not save_dc_state(obj.hw, filename):
                            obj.error("Failed to save %s state" %
                                      DC8051_STATE_NAME)
                            return False
                        output.append("%s state saved to %s" %
                                      (DC8051_STATE_NAME, filename))
                        return False
                    obj.error("Invalid state '%s'" % state)
                    return False
                state = HwStates.byname(state.upper())
            else: state, filename = HwStates.HW, words[1]
            # expand environment variables in filename
            filename = os.path.expanduser(os.path.expandvars(filename))
            try:
                if state == HwStates.HW:
                    # HW state needs to be captured first
                    if obj.hw.has_secure():
                        obj.show_cmd_output(JTAG_WARNING)
                    obj.hw.capture_state()
                if obj.hw.save_state(filename, state):
                    output.append("Saved state%sto %s" % \
                                  (state and " '%s' " % state or "",
                                   filename))
            except HwError, e:
                obj.hw_error(e)
                return False
        else:
            filename = os.path.expanduser(os.path.expandvars(words[1]))
            ret = obj.hw.load_state(filename)
            if ret is False:
                obj.error("Failed to load HW state from %s" % filename)
            else:
                output.append("Loaded HW state from %s in '%s'" % \
                              (filename, "%s" % ret))
    elif words[0] == "switch":
        if len(words) != 2:
            obj.error("Invalid argument list for 'state swtich' command")
            return False
        try:
            new = obj.hw.switch_state(words[1])
            output.append("New active state is %s" % new)
            if new == HwStates.HW: obj.set_prompt()
            else: obj.set_prompt(str(new))
        except HwError, e:
            obj.hw_error(e)
            return False
    elif words[0] == "capture":
        if obj.hw.has_secure():
            obj.show_cmd_output(JTAG_WARNING)
        try: obj.hw.capture_state()
        except HwError, e:
            obj.hw_error(e)
            return False
    elif words[0] == "diff":
        if len(words) != 3:
            obj.error("Diff command requires two states to compare")
            return False
        states = words[1:]
        for state in states:
            if state.upper() not in HwStates:
                obj.error("Unknown state '%s'" % state)
                return False
        # convert to the enums
        states = [HwStates.byname(x.upper()) for x in states]
        for state in [x for x in states \
                      if x not in (HwStates.HW, HwStates.CAPTURED)]:
            if not obj.hw.have_saved_state(state):
                obj.error("No stored state '%s'" % state)
                return False
        if HwStates.HW in states:
            if obj.hw.has_secure():
                obj.show_cmd_output(JTAG_WARNING)
            try: obj.hw.capture_state()
            except HwError, e:
                obj.hw_error(e)
                return False
            # change HW to CAPTURED after we've captured the state
            states = [HwStates.CAPTURED if x == HwStates.HW else x for x in states]

        obj.show_cmd_output([("Register", "%18s => %18s" % \
                              (str(states[0]) + " Value",
                               str(states[1]) + " Value")), SPACER])

        def show_diff(reg, ctxt, idx, val1, val2):
            obj.show_cmd_output([("%s[%d][%d]" % (reg.name, ctxt, idx),
                                  "%#18x => %#18x" % (val1, val2))], False)
        obj.hw.diff_states(states[0], states[1], show_diff)
    elif words[0] == "show":
        states = obj.hw.describe_states()
        if states:
            output += [("%-10s   %-19s" % ("STATE", "CAPTURED"), "FROM"), SPACER]
            for state in states:
                output.append(("%-10s   %-19s" % (state[0], formattime(state[1])),
                               state[2]))
        else: output.append("No HW state information")
    return False

PCI_COMMANDS = ["read", "write"]

def pci_config_ops(obj, line, register_output):
    words = line.split()

    if not len(words):
        obj.error("Incomplete command")
        return False

    if words[0] not in PCI_COMMANDS:
        obj.error("Unknown argument '%s' for 'pci' command" % words[0])
        return False

    cmd = words[0]
    value = None
    if cmd == "write":
        if len(words) < 3:
            obj.error("Not enough arguments for '%s' command" % cmd)
            return False
        else:
            if is_number(words[2]):
                if is_hex(words[2]): value = c_ulonglong(int(words[2], 16))
                else: value = c_ulonglong(int(words[2]))
            else:
                obj.error("Invalid write value '%s'" % words[2])
                return False

    reginfo = []
    if not validate_register_spec(words[1], reginfo, obj.hw,
                                  obj.hw.get_config_register_names()):
        obj.error(reginfo[0])
        return False

    for csr in reginfo:
        reg, cs, ce, rs, re, field = csr
        for idx in range(cs, ce+1):
            for elem in range(rs, re+1):
                try:
                    if cmd == "read":
                        result = obj.hw.read_config_register(reg, idx, elem)
                        addr = obj.hw.addressof(reg, idx, elem)
                        if obj.config.show_register_bitfields:
                            decoded = obj.hw.decode_register(reg, result)
                            if field: show_fields = [getattr(decoded, field)]
                            else: show_fields = decoded.fields
                        else: show_fields = None
                        freg = obj.format_register(reg, idx, elem, addr, result,
                                                   show_fields)
                        if freg: register_output.append(freg + "\n")
                    elif cmd == "write":
                        if field:
                            obj.hw.write_config_register_field(reg, idx, elem, field, value)
                        else:
                            obj.hw.write_config_register(reg, idx, elem, value)
                except HwError, e:
                    obj.hw_error(str(e).strip())
                    break
    return False

def pci_config_comp(obj, text, line, start_index, end_index):
    words = line.split()
    if len(words) == 1: return PCI_COMMANDS
    elif len(words) == 2:
        if text: return find_matches(text, PCI_COMMANDS)
        else: return obj.hw.get_config_register_names()
    elif len(words) == 3 and text:
        return register_name_complete(obj.hw, text, line, start_index, end_index,
                                      obj.hw.get_config_register_names())
    return []

def pci_config_help():
    return {"synopsis": "pci read|write <PCI register> [<value>]",
            "description":
            "Read or write from/to a PCI configuration space register",
            "arguments":
            ["Both the 'read' and 'write' variable require the " + \
             "'<PCI register>' argument. This argument is a symbolic " + \
             "name of a register within the PCI(s) configuration space. " + \
             "\n\nThe 'read' variant of the command " + \
             "will read the value of a 32bit PCI(e) configuration register." + \
             "\n\nThe 'write' variant will write '<value>' to the PCI(e) " + \
             "configuration space register.", "",
             [('<PCI register>', "The name of the PCI(e) configuration space " + \
               "register to operate on. Name used follow register names as " + \
               "defined by the HAS. This argument supports all the normal " + \
               "addressing modes and syntax as with other registers."),
              ('<value>', "This argument is unused when reading a register and " + \
               "is required when writing a register. It specifies the numerical " + \
               "value (in hexidecimal format) to write to the register.")]
             ],
            "returns":
            "The 'read' variant will return the decoded value of the PCI(e) " + \
            "configuration space registers. The 'write' variant does not " + \
            "return a value.",
            "see also": "Addressing, read, write, decode"}

def save_dc_state(hw, filename):
    try:
        with open(dc8051_mem_file(hw), 'rb', 0) as fd:
            buf = fd.read(DC8051_DATA_MEM_SIZE)
        with open(filename, 'wb') as fd:
            fd.write(buf)
    except IOError:
        return False
    return True


__commands__ = [("status", describe_hfi, None, describe_help),
                ("state", state_save, complete_state, help_state),
                ("pci", pci_config_ops, pci_config_comp, pci_config_help)]
