# INTEL CONFIDENTIAL
# Copyright 2014 2015 Intel Corporation All Rights Reserved.
#
# The source code contained or described herein and all
# documents related to the source code ("Material") are
# owned by Intel Corporation or its suppliers or licensors.
# Title to the Material remains with Intel Corporation or
# its suppliers and licensors. The Material may contain trade
# secrets and proprietary and confidential information of Intel
# Corporation and its suppliers and licensors, and is protected
# by worldwide copyright and trade secret laws and treaty
# provisions. No part of the Material may be used, copied,
# reproduced, modified, published, uploaded, posted, transmitted,
# distributed, or disclosed in any way without Intel's prior express
# written permission.
#
# No license under any patent, copyright, trade secret or other
# intellectual property right is granted to or conferred upon you
# by disclosure or delivery of the Materials, either expressly, by
# implication, inducement, estoppel or otherwise. Any license under
# such intellectual property rights must be express and approved by
# Intel in writing.
#
# Unless otherwise agreed by Intel in writing, you may not remove or
# alter this notice or any other notice embedded in Materials by Intel
# or Intel's suppliers or licensors in any way.
from lib.diags import *
import os

QWORD_LEN = 8 # in bytes
DWORD_LEN = 4 # in bytes
DEFAULT_WORD_SIZE = DWORD_LEN
BYTES_PER_LINE = 16
PAGE_SHIFT = 12
PAGE_MASK = (~((1 << PAGE_SHIFT) - 1))

def virt2phys(addr, pid):
    pagemap = "/proc/%d/pagemap" % pid
    vpn = addr >> PAGE_SHIFT
    offset = addr & ~PAGE_MASK

    # Exceptions are caught by the caller
    with open(pagemap , "rb") as pfd:
        pfd.seek(vpn * QWORD_LEN, os.SEEK_SET)
        page_val = pfd.read(QWORD_LEN)

    value, shift = 0, 0
    for byte in page_val:
        value |= (ord(byte) << shift)
        shift += QWORD_LEN

    present = (value >> 63) & 0x1
    pfn = value & ((1 << 55) - 1)

    if not present:
        raise ValueError("Page not present")
    return (pfn << PAGE_SHIFT) + offset

def mempeek_help():
    return {"synopsis": "mempeek <type> [<pid>] <address> [<length>]",
            "description":
            "The 'mempeek' command reads data from host memory. The " + \
            "command can handle both physical and virtual addresses.\n\n" + \
            "Since the command can read any physical address, it requires " + \
            "that hfidiags be run with superuser privileges.\n\n" + \
            "When attempting to read a virtual address, the command " + \
            "requires the PID of the process who's virtual address " + \
            "is being read.",
            "arguments" :
            ["The argument list of the 'mempeek' command depends on the " + \
             "type of the address to be read:", "",
             [("<type>", "Used to specify what type of address is being used - " + \
               "'phys' for physical, 'virt' for virtual. This is a required " + \
               "argument."),
              ("<pid>", "When using a virtual address, the command " + \
               "needs the PID of the process to whom the address belongs in " + \
               "order to properly find the data."),
              ("<address>", "The starting address of the memory to be read."),
              ("<length>", "The amount of memory to read (in bytes). If " + \
               "this argument is omitted, 8 bytes will be read.")]],
            "returns" :
            "If memory is successfully read, the command will display " + \
            "the data. Otherwise, an error message is displayed."
            }

def mempeek_complete(obj, text, line, start_index, end_index):
    cmdline = line.split()
    if len(cmdline) >= 2 and not text: return []
    addr_types = ["phys", "virt"]
    if text: return find_matches(text, addr_types)
    return addr_types

def mempeek(obj, line, output):
    words = line.split()

    # Have to run as 'root' to allow reading of memory
    if os.geteuid() != 0:
        obj.error("Command can only be used if running with superuser privileges")
        return False

    if not words:
        obj.error("Missing required command arguments")
        return False

    if len(words) < 2:
        obj.error("Too few arguments")
        return False

    addr, paddr, pid, length = 0, 0, -1, DEFAULT_WORD_SIZE
    addr_str, pid_str, length_str = "", "", ""

    addrtype = words[0]
    if len(addrtype) > 4: addrtype = addrtype[:4]
    if addrtype == "phys":
        if len(words[1:]) < 1:
            obj.error("At least address argument is required")
            return False
        addr_str = words[1]
        if len(words[1:]) > 1:
            length_str = words[2]
    elif addrtype == "virt":
        if len(words[1:]) < 2:
            obj.error("ProcessID and virtual address arguments required")
            return False
        pid_str = words[1]
        addr_str = words[2]
        if len(words[1:]) > 2:
            length_str = words[3]
    else:
        obj.error("Address types can only be 'virt' or 'phys'")
        return False

    if not is_hex(addr_str):
        obj.error("Memory address argument required to be HEX value")
        return False
    else:
        addr = int(addr_str, 16)

    if addrtype == "virt":
        if not is_number(pid_str):
            obj.error("ProcessID arugment must be a numerical value")
            return False
        else:
            pid = int(pid_str, is_hex(pid_str) and 16 or 10)

    if length_str:
        if not is_number(length_str):
            obj.error("Length argument must be a numerical value")
            return False
        else:
            length = int(length_str, is_hex(length_str) and 16 or 10)

    if addrtype == "virt":
        try: paddr = virt2phys(addr, pid)
        except:
            obj.error("Failed to translate virtual address")
            return False
    else: paddr = addr

    try:
        with open("/dev/mem", 'rb') as devmem:
            devmem.seek(paddr, os.SEEK_SET)
            bytes = devmem.read(length)
    except IOError, e:
        obj.error(str(e))
        return False

    if bytes:
        line, vfmt = ["0x%012x:" % addr], "0x%%0%dx" % (DEFAULT_WORD_SIZE * 2)
        newline = 0
        for start in range(0, length, DEFAULT_WORD_SIZE):
            qword = bytes[start:start+DEFAULT_WORD_SIZE]
            value, shift = 0, 0
            for idx in range(DEFAULT_WORD_SIZE):
                value |= int((ord(qword[idx]) << shift))
                shift += 8
            line.append(vfmt % value)
            addr += DEFAULT_WORD_SIZE
            newline += 1
            if newline == BYTES_PER_LINE / DEFAULT_WORD_SIZE:
                output.append(" ".join(line))
                line = ["0x%012x:" % addr]
                newline = 0
        if newline: output.append(" ".join(line))
    else:
        obj.error("no bytes read")
        return False

    return False

__commands__ = [("mempeek", mempeek, mempeek_complete, mempeek_help)]
