# INTEL CONFIDENTIAL
# Copyright 2014 2015 Intel Corporation All Rights Reserved.
#
# The source code contained or described herein and all
# documents related to the source code ("Material") are
# owned by Intel Corporation or its suppliers or licensors.
# Title to the Material remains with Intel Corporation or
# its suppliers and licensors. The Material may contain trade
# secrets and proprietary and confidential information of Intel
# Corporation and its suppliers and licensors, and is protected
# by worldwide copyright and trade secret laws and treaty
# provisions. No part of the Material may be used, copied,
# reproduced, modified, published, uploaded, posted, transmitted,
# distributed, or disclosed in any way without Intel's prior express
# written permission.
#
# No license under any patent, copyright, trade secret or other
# intellectual property right is granted to or conferred upon you
# by disclosure or delivery of the Materials, either expressly, by
# implication, inducement, estoppel or otherwise. Any license under
# such intellectual property rights must be express and approved by
# Intel in writing.
#
# Unless otherwise agreed by Intel in writing, you may not remove or
# alter this notice or any other notice embedded in Materials by Intel
# or Intel's suppliers or licensors in any way.
from lib.diags import *
from lib.hw import HwError
from ctypes import *
import time
import sys

def clear_screen():
    sys.stderr.write("\x1b[2J\x1b[H")

def watch_cmd_complete(obj, text, line, start_index, end_index):
    return register_name_complete(obj.hw, text, line, start_index, end_index,
                                  obj.hw.get_register_names())

def watch_help():
    return {"synopsis": "watch <register spec> | <addr> [<register spec> | " + \
            "<addr>] [<timeout>]",
            "description": "Continously monitor CSRs for changes. This " + \
            "command will keep reading the values of all CSRs or addresses " + \
            "specified on the command line and display the current values.",
            "arguments":
            "The command accepts both symbolic register specifications (see " + \
            "\"Addressing\" in the help) and HW memory addresses.\n\n" + \
            "Multiple CSRs or addresses can be monitored at the same time by " + \
            "specifing them on the command line.\n\nThe last command line " + \
            "option is a floating point number representing the interval at " + \
            "which to read the values (default is 1 second).\n\nNOTE: Since " + \
            "there is no reason to monitor CSRs from a saved state, the " + \
            "current active state will be ignored and the command will " + \
            "monitor only HW CSRs.",
            "returns": "N/A",
            "see also": "Addressing, read"}

def filter_watch_csrs(csrlist):
    filtered = []
    for csr in csrlist:
        name, cs, ce, rs, re, field = csr
        if name not in [x[0] for x in filtered]: filtered.append(csr)
        else:
            f = [x for x in filtered if x[0] == name][0]
            idx = filtered.index(f)
            if f[5] == None: filtered[idx][5] = []
            elif isinstance(f[5], str): filtered[idx][5] = [f[5]]
            # Turn the ranges into sets so we can use the 'set' object
            # methods
            fset, nset = set(range(f[1], f[2]+1)), set(range(cs, ce+1))
            if fset.isdisjoint(nset):
                # Disjoint sets of the same CSR are considered different
                filtered.append(csr)
                continue
            else:
                nset = sorted(list(fset.union(nset)))
                # The UI only supports consequitive, non-disjoint range
                # so we can do the assignment below and not worry that
                # we are missing CSRs.
                filtered[idx][1], filtered[idx][2] = nset[0], nset[-1]
            fset, nset = set(range(f[3], f[4]+1)), set(range(rs, re+1))
            if fset.isdisjoint(nset):
                filtered.append(csr)
                continue
            else:
                nset = sorted(list(fset.union(nset)))
                # The UI only supports consequitive, non-disjoint range
                # so we can do the assignment below and not worry that
                # we are missing CSRs.
                filtered[idx][3], filtered[idx][4] = nset[0], nset[-1]
            if field and field not in filtered[idx][5]:
                filtered[idx][5].append(field)
    return filtered

def csr_watch(obj, line, output):
    interval = 1.0 # second
    if not line:
        obj.error("Invalid watch command")
        return False
    words = line.split()

    if is_number(words[-1]) and not is_hex(words[-1]):
        interval = float(words[-1])
        words = words[:-1]

    if not words:
        obj.error("No CSRs or addresses to monitor")
        return False

    csrlist = []
    addrlist = []
    for word in words:
        if is_hex(word):
            addr = c_ulonglong(int(word, 16))
            if addr.value not in [x.value for x in addrlist]: addrlist.append(addr)
        else:
            reginfo = []
            if not validate_register_spec(word, reginfo, obj.hw): continue
            csrlist += reginfo

    csrlist = filter_watch_csrs(csrlist)
    numlines = obj.screen_height()

    while True:
        clear_screen()
        data = [("Continous CSR monitor", "Interval: %.2f second(s)" % interval),
                "(use Ctrl-C to stop)", SPACER, ""]
        for addr in addrlist:
                result = obj.hw.read(c_ulonglong(int(word, 16)))[0]
                _v = obj.format_numeric_value(result)
                if _v: data.append((word, _v))
        for csr in csrlist:
                reg, cs, ce, rs, re, fields = csr
                for idx in range(cs, ce+1):
                    for elem in range(rs, re+1):
                        try: result = obj.hw.read_register(reg, idx, elem, "hw")
                        except HwError, e:
                            data.append(("%s[%d][%d]" % (reg, idx, elem), str(e)))
                            break

                        if fields:
                            if isinstance(fields, str): fields = [fields]
                            decoded = obj.hw.decode_register(reg, result, fields)
                            for field in fields:
                                field = getattr(decoded, field)
                                _v = obj.format_numeric_value(field.value,
                                                              field.width, True)
                                if _v:
                                    data.append(("%s[%d][%d].%s" % (reg, idx, elem,
                                                                    field.name), _v))

                        else:
                            _v = obj.format_numeric_value(result)
                            if _v: data.append(("%s[%d][%d]" % (reg, idx, elem), _v))

        obj.show_cmd_output(data[:numlines-1], False, paging=False)
        if interval: time.sleep(interval)

    return False

__commands__ = [("watch", csr_watch, watch_cmd_complete, watch_help)]

