# INTEL CONFIDENTIAL
# Copyright 2014 2015 Intel Corporation All Rights Reserved.
#
# The source code contained or described herein and all
# documents related to the source code ("Material") are
# owned by Intel Corporation or its suppliers or licensors.
# Title to the Material remains with Intel Corporation or
# its suppliers and licensors. The Material may contain trade
# secrets and proprietary and confidential information of Intel
# Corporation and its suppliers and licensors, and is protected
# by worldwide copyright and trade secret laws and treaty
# provisions. No part of the Material may be used, copied,
# reproduced, modified, published, uploaded, posted, transmitted,
# distributed, or disclosed in any way without Intel's prior express
# written permission.
#
# No license under any patent, copyright, trade secret or other
# intellectual property right is granted to or conferred upon you
# by disclosure or delivery of the Materials, either expressly, by
# implication, inducement, estoppel or otherwise. Any license under
# such intellectual property rights must be express and approved by
# Intel in writing.
#
# Unless otherwise agreed by Intel in writing, you may not remove or
# alter this notice or any other notice embedded in Materials by Intel
# or Intel's suppliers or licensors in any way.
import re
import os
import sys
import time
from ctypes import *

SPACER     = (1<<0)
HEADER     = (1<<1)
EMPTY_LINE = (1<<2)
regexp_chars = ('*', '?', '+', '{')

class Progress:
    progresschars = "|/-\\"
    index = 0
    def move(self):
        sys.stdout.write("\b%s" % self.progresschars[self.index])
        sys.stdout.flush()
        self.index = (self.index + 1) % len(self.progresschars)

def is_hex(value):
    """Evaluate whether a value is a hexidecimal number. Return True if so,
    False otherwise."""
    try:
        if value[:2] in ("0x", "0X"):
            int(value,16)
            return True
    except ValueError: pass
    return False

def is_float(value):
    try:
        float(value)
        return True
    except ValueError: pass
    return False

def is_number(value):
    """Evaluate whether a value is a number (decimal or hexidecimal. Return
    True if so, False otherwise."""
    if is_hex(value) or is_float(value): return True
    return isinstance(value, str) and value.isdigit()

def str2num(str):
    """Convert a string value to a number. The string value can be a decimal
    or hexidecimal representation of a number. Returns None if the string
    is not a numerical value."""
    if is_number(str):
        if is_hex(str): return int(str, 16)
        else: return int(str)
    return None

def is_regexp(buffer):
    return True in set([x in buffer for x in regexp_chars])

def find_matches(text, allmatches, unique=False, match_flags=re.I):
    """Find all strings that are either matches for could be matches for
    text. This function is mostly used to find matches for register names.
    If unique is True, the function returns a list of strings from allmatches
    that fully match text (text == allmatches[i]). If unique is False (or
    unique is True but no complete matches have been found), the function
    returns a list of possible matches (the string in allmatches begins with
    'text'). If not matches are found, an empty list is returned."""
    matches = []
    if text:
        text = '^' + text
        if is_regexp(text):
            for char in regexp_chars:
                text = text.replace(char, '.'+char)
        if unique: text += '$'
        regexp = re.compile(text, match_flags)
        if unique: matches = filter(regexp.match, allmatches)
        else: matches = filter(regexp.search, allmatches)
    else:
        matches = allmatches
    return matches

def parse_register_spec(spec):
    """Parse a register command line input. On the command line, the user
    can specify a register name in one of the following ways: name, name.field,
    name[index], name[start-end], name[index].field, name[start-end].field.
    This function parses the input and returns a tuple with register name,
    starting index, ending index, and field (in that order). In the case where
    no starting or ending index are present, the function uses the values '0'
    and '-1' for starting and ending index respectively."""
    def parse_range(r):
        if not r: return (0, -1)
        if r == '*': rs, re = "0", "-1"
        elif '-' in r: rs, re = r.split('-', 1)
        else: rs, re = r, r
        try: rs, re = int(rs), int(re)
        except (ValueError, TypeError): rs, re = -1, -1
        return (rs, re)
    regname, ctxts, regrange, field = None, None, None, None
    if spec[-1] == '.': spec = spec[:-1]
    if '.' in spec: reg, field = spec.split('.', 1)
    else: reg, field = spec, None
    ps = reg.split('[')
    if len(ps) >= 1: reg, ctxts, regrange = ps[0], "*", "*"
    if len(ps) >= 2 and ps[1]: ctxts = ps[1][-1] == ']' and ps[1][:-1] or "None"
    if len(ps) == 3 and ps[2]: regrange = ps[2][-1] == ']' and ps[2][:-1] or "None"
    elif len(ps) > 3: ctxts, regrange = "None", "None"
    ctxts = parse_range(ctxts)
    regrange = parse_range(regrange)
    return (reg,) + ctxts + regrange + (field,)

def path_name_complete(text, line, startidx, endidx):
    """Generic filename completion function."""
    def annotate(l, p="."):
        return map(lambda x: os.path.isdir(p+os.sep+x) and x+os.sep or x, l)
    if not text: return annotate(os.listdir("."))
    else:
        # expand all variables first
        text = os.path.expandvars(text)
        dirname = os.path.expanduser(os.path.dirname(text))
        filename = os.path.basename(text)
        if not dirname: return find_matches(text, annotate(os.listdir(".")))
        else:
            if os.path.exists(dirname) and os.path.isdir(dirname):
                dirlist = annotate(os.listdir(dirname), dirname)
                if filename:
                    return [(dirname.endswith(os.sep) and dirname or \
                             dirname+os.sep) + x for x in \
                            find_matches(filename, dirlist)]
                else: return [(dirname.endswith(os.sep) and dirname or \
                               dirname+os.sep) + x for x in dirlist]
    return []

def register_name_complete(hw, text, line, startidx, endidx, all=None):
    """Generic completion function which returns a list with all possible
    registers matching 'text'. The function handles any of the formats
    supported by the UI interface for register naming and indexing."""
    if all == None:
        all = hw.get_register_names() + hw.get_config_register_names()
    cmdline = line.split()
    if text:
        if "." in text:
            regprefix, field = text.split('.', 1)
            if '[' in regprefix: reg = regprefix[:regprefix.find('[')]
            else: reg = regprefix
            fields = []
            for r in find_matches(reg, all, not is_regexp(reg)):
                regprefix = regprefix.replace(regprefix[0:len(reg)-1],
                                              r[0:len(reg)-1])
                fields += hw.get_register_fields(r)
            if field:
                matches = find_matches(field, fields)
                if len(matches) == 1: return [regprefix + "." + matches[0]]
                else: return [regprefix + '.' + x for x in matches]
            else: return [regprefix + '.' + x for x in fields]
        elif '[' in text:
            reg, idx = text.split('[', 1)
            idxes = hw.get_register_index_names()
            if idxes: return [reg+'['+x+']' for x in find_matches(idx, idxes)]
            else: return []
        else: return find_matches(text, all)
    else: return all

def validate_register_spec(spec, output, hw=None, namespace=None):
    """Given a register specification (normally, from the command line),
    validate its format by checking for valid register names, valid
    register indeces, and valid indeces ranges. On error, the function
    puts an appropriate error message as the first element of the 'output'
    array and return False. On success, the output array contains the
    parsed, validated (and potentially, expanded) register spec and the
    function return True."""
    if hw:
        # Perform any register index substitutions that are needed
        indexes = hw.get_register_index_names()
        for index in indexes:
            if index in spec:
                spec = spec.replace(index, str(hw.get_register_index(index)))
    reg, cs, ce, rs, re, field = parse_register_spec(spec)
    if not reg or (cs == -1 and rs == -1):
        output.append("Malformed register specification")
        return False
    if cs == -1:
        output.append("Invalid context index")
        return False
    if rs == -1:
        output.append("Invalid register array index")
        return False
    if (ce != -1 and cs > ce) or (re != -1 and rs > re):
        output.append("Invalid index range (start is larger than end)")
        return False
    ret = [reg, cs, ce, rs, re, field]
    if hw:
        if not namespace:
            namespace = hw.get_register_names()
        matches = find_matches(ret[0], namespace, True)
        if not matches:
            if not is_regexp(ret[0]):
                output.append("Unknown register '%s'" % ret[0])
            else:
                output.append("Pattern '%s' does not match any CSR name" % ret[0])
            return False
        elif len(matches) > 1:
            if not is_regexp(ret[0]):
                output.append("Ambiguous register name '%s'. " + \
                              "Possible options are:\n%s" % \
                              (reg, ", ".join(matches)))
                return False
        reglist = matches
        specinfo = ret[:]
        ret = []
        for reg in reglist:
            add, reginfo = True, specinfo[:]
            reginfo[0] = reg
            rcount = hw.get_register_count(reg)
            asize = hw.get_register_array_size(reg)
            if rcount == 1 and asize > 1 and not specinfo[3]:
                reginfo[3], reginfo[4] = specinfo[1], specinfo[2]
                reginfo[1], reginfo[2] = 0, 0
            if reginfo[2] == -1: reginfo[2] = rcount - 1
            elif reginfo[2] >= rcount:
                output.append("Register '%s' has only %d instance%s." % \
                              (reg, rcount, rcount > 1 and "s (0-%d)" % (rcount-1) \
                               or ""))
                return False
            if reginfo[4] == -1: reginfo[4] = asize - 1
            elif reginfo[4] >= asize:
                output.append("Register '%s' has only %d array element%s." % \
                              (reg, asize, asize > 1 and "s (0-%d)" % (asize-1) or ""))
                return False
            if reginfo[5]:
                regfields = find_matches(reginfo[5], hw.get_register_fields(reg), True)
                if regfields:
                    reginfo[5] = regfields[0]
                else:
                    add = False
            if add:
                ret.append(reginfo)
    output += ret
    return True

def formattime(timestamp=time.time(), format="%Y-%m-%d %H:%M:%S"):
    if isinstance(timestamp, (float, int)):
        timestamp = time.localtime(timestamp)
    elif isinstance(timestamp, tuple) and len(timestamp) == 9:
        timestamp = time.struct_time(timestamp)
    if not isinstance(timestamp, time.struct_time): return "None"
    return time.strftime(format, timestamp)

def make_python_version(major, minor, release):
    return ((major & 0xFF) << 24) | ((minor & 0xFF) << 16) | \
           ((release & 0xFF) << 8) | 0xF0
def py_version_cmp(major, minor, release):
    return cmp(sys.hexversion, make_python_version(major, minor, release))
