# INTEL CONFIDENTIAL
# Copyright 2014 2015 Intel Corporation All Rights Reserved.
#
# The source code contained or described herein and all
# documents related to the source code ("Material") are
# owned by Intel Corporation or its suppliers or licensors.
# Title to the Material remains with Intel Corporation or
# its suppliers and licensors. The Material may contain trade
# secrets and proprietary and confidential information of Intel
# Corporation and its suppliers and licensors, and is protected
# by worldwide copyright and trade secret laws and treaty
# provisions. No part of the Material may be used, copied,
# reproduced, modified, published, uploaded, posted, transmitted,
# distributed, or disclosed in any way without Intel's prior express
# written permission.
#
# No license under any patent, copyright, trade secret or other
# intellectual property right is granted to or conferred upon you
# by disclosure or delivery of the Materials, either expressly, by
# implication, inducement, estoppel or otherwise. Any license under
# such intellectual property rights must be express and approved by
# Intel in writing.
#
# Unless otherwise agreed by Intel in writing, you may not remove or
# alter this notice or any other notice embedded in Materials by Intel
# or Intel's suppliers or licensors in any way.
import exceptions

class EnumError(exceptions.Exception): pass

class EnumType(object):
    """Object type for values in enumeration sets. The difference
    between this type and any othe Python object type is that it
    can not be setup more than once. This is done in order to best
    emulate C's enumeration types."""
    def __init__(self, name, value):
        self.__value, self._name = value, name
    def get(self): return self.__value
    def __setattr__(self, attr, value):
        if attr in self.__dict__: raise TypeError("Can not change enum value")
        self.__dict__[attr] = value
    def __eq__(self, other):
        if isinstance(other, self.__class__): return self.get() == other.get()
        return False
    def __compute(self, other, op=None, rev=False):
        if not op: raise TypeError("Unspecified arithmetic operation")
        if isinstance(other, EnumType): arg = other.__value
        elif isinstance(other, int): arg = other
        else:
            raise TypeError("unsupported operand type(s) for %s: '%s' and '%s'" % \
                            (op, self.__class__.__name__, other.__class__.__name__))
        if rev: return eval("%d %s %d" % (arg, op, self.__value))
        else: return eval("%d %s %d" % (self.__value, op, arg))
    def __str__(self): return self._name
    def __repr__(self): return "Enum[%s](%d)" % (self._name, self.__value)
    def __hash__(self): return hash(self._name) + hash(self.__value)
    def __cmp__(self, other): return cmp(self.__value, other.__value)
    def __add__(self, other): return self.__compute(other, "+")
    def __sub__(self, other): return self.__compute(other, "-")
    def __mul__(self, other): return self.__compute(other, "*")
    def __div__(self, other): return self.__compute(other, "/")
    def __pow__(self, other): return self.__compute(other, "**")
    def __and__(self, other): return self.__compute(other, "&")
    def __or__(self, other): return self.__compute(other, "|")
    def __xor__(self, other): return self.__compute(other, "^")
    def __radd__(self, other): return self.__compute(other, "+", True)
    def __rsub__(self, other): return self.__compute(other, "-", True)
    def __rmul__(self, other): return self.__compute(other, "*", True)
    def __rdiv__(self, other): return self.__compute(other, "/", True)
    def __rpow__(self, other): return self.__compute(other, "**", True)
    def __rand__(self, other): return self.__compute(other, "&", True)
    def __ror__(self, other): return self.__compute(other, "|", True)
    def __rxor__(self, other): return self.__compute(other, "^", True)
    def __int__(self): return self.__value
    def __long__(self): return long(self.__value)
    def __float__(self): return float(self.__value)
    def __oct__(self): return oct(self.__value)
    def __hex__(self): return hex(self.__value)
    def __coerce__(self, other): return (self.__value, int(other))

class Enum(object):
    """The Enum object is meant to emulate C's enumerations. As such,
    its members are set only once and any further attempt to alter the
    enumaration set will raise an exception."""
    __initted__ = False
    def __init__(self, members, startidx=0, values=None):
        if values != None:
            if len(members) != len(values):
                raise EnumError("sizeof(members) != sizeof(values)")
            if False in set([isinstance(x, (int, long)) for x in values]):
                raise EnumError("non-integer values not supported")
            self._members_ = frozenset([EnumType(members[idx], values[idx]) \
                                        for idx in range(len(members))])
        else:
            self._members_ = frozenset([EnumType(members[idx], startidx+idx) \
                                        for idx in range(len(members))])
        self.idx = 0
        self.__start = startidx
        self.__initted__ = True
    def __getattr__(self, attr):
        if attr in self:
            return filter(lambda x: attr == x._name, self._members_)[0]
        raise AttributeError("Enum does not contain member '%s'" % attr)
    def __setattr__(self, attr, value):
        if self.__initted__: raise TypeError("Can not change enum set")
        super(Enum, self).__setattr__(attr, value)
    def __contains__(self, enum):
        if isinstance(enum, EnumType):
            return True in set([enum == x for x in self._members_])
        elif isinstance(enum, (int, long)):
            return enum in [x.get() for x in self._members_]
        elif isinstance(enum, str):
            return enum in [x._name for x in self._members_]
        return False
    def __str__(self):
        __m = sorted(self._members_)
        return "Enum("+",".join(["%s=%d"%(x._name,int(x)) for x in __m])+")"
    def __eq__(self, other):
        if sorted(self._members_) != sorted(other._members_):
            return False
        return True
    def __iter__(self):
        for member in sorted(self._members_):
            yield member
    def __len__(self):
        return len(self._members_)
    def __add__(self, other):
        if other not in self:
            if isinstance(other, EnumType): s = self._members_ | set([other])
            elif isinstance(other, str):
                s = self._members_ | set([EnumType(other,
                                                   self.__start + len(self))])
            elif isinstance(other, int):
                raise AttributeError("Cannot add EnumType without name")
            elif isinstance(other, Enum):
                s = self._members_ | other._members_
            return Enum([str(x) for x in s], values=[int(x) for x in s])
    def __sub__(self, other):
        if isinstance(other, EnumType): s = self._members_ - set([other])
        elif isinstance(other, str): s = self._members_ - set([getattr(self, other)])
        elif isinstance(other, int):
            s = self._members_ - set(filter(lambda x: int(x) == other,
                                            self._members_))
        elif isinstance(other, Enum): s = self._members_ - other._members_
        return Enum([str(x) for x in s], values=[int(x) for x in s])
    def __and__(self, other):
        if not isinstance(other, Enum):
            raise TypeError("invalid type '%s' for & operation" % \
                            other.__class__.__name__)
        s = self._members_ & other._members_
        return Enum([str(x) for x in s], values=[int(x) for x in s])
    def __xor__(self, other):
        if not isinstance(other, Enum):
            raise TypeError("invalid type '%s' for ^ operation" % \
                            other.__class__.__name__)
        s = self._members_ ^ other._members_
        return Enum([str(x) for x in s], values=[int(x) for x in s])
    def first(self):
        return tuple(sorted(self._members_))[0]
    def last(self):
        return tuple(sorted(self._members_))[-1]
    def next(self):
        self.idx = self.idx + 1 if self.idx < len(self._members_) \
                   else self.idx
        return tuple(sorted(self._members_))[self.idx - 1]
    def byval(self, value):
        if value in self and isinstance(value, (int, long)):
            return filter(lambda x: value == x.get(), self._members_)[0]
        raise AttributeError("Enum does not contain value '%s'" % value)
    def byname(self, value):
        if isinstance(value, str) and value in self:
            return filter(lambda x: value == x._name, self._members_)[0]
        raise AttributeError("Enum does not contain value '%s'" % value)
