# INTEL CONFIDENTIAL
# Copyright 2014-2018 Intel Corporation All Rights Reserved.
#
# The source code contained or described herein and all
# documents related to the source code ("Material") are
# owned by Intel Corporation or its suppliers or licensors.
# Title to the Material remains with Intel Corporation or
# its suppliers and licensors. The Material may contain trade
# secrets and proprietary and confidential information of Intel
# Corporation and its suppliers and licensors, and is protected
# by worldwide copyright and trade secret laws and treaty
# provisions. No part of the Material may be used, copied,
# reproduced, modified, published, uploaded, posted, transmitted,
# distributed, or disclosed in any way without Intel's prior express
# written permission.
#
# No license under any patent, copyright, trade secret or other
# intellectual property right is granted to or conferred upon you
# by disclosure or delivery of the Materials, either expressly, by
# implication, inducement, estoppel or otherwise. Any license under
# such intellectual property rights must be express and approved by
# Intel in writing.
#
# Unless otherwise agreed by Intel in writing, you may not remove or
# alter this notice or any other notice embedded in Materials by Intel
# or Intel's suppliers or licensors in any way.
import os
import re
import io
import bz2
import mmap as mod_mmap
import time
import cPickle
import StringIO
import struct
import hashlib
import exceptions
import lib.diags
import lib.enums
import lib.pci
import ctypes.util as util
from ctypes import *
from copy import deepcopy
from lib.logger import *
import collections

ADDR_INVALID = 0xFFFFFFFFFFFFFFFF
DEV_DIR_PATH = "/dev"

# HW DB load error codes
DB_ERR_VER          = -1
DB_ERR_UNPACK       = -2
DB_ERR_FORMAT       = -3
DB_ERR_HWREV        = -4
DB_ERR_HWREV_UNSUP  = -5
DB_ERR_NO_DB_LOADED = -6

DbFileFormats = lib.enums.Enum(["FLAT", "BZIP"])

# register flags
RF_NON_READABLE    = (1 << 0)  # register is non-readable
RF_LOCKED          = (1 << 1)  # register has a "locked" resource handler

REG_FIELD_TYPE_RO  = (1 << 1)
REG_FIELD_TYPE_RW  = (1 << 2)
REG_FIELD_TYPE_WO  = (1 << 3)
REG_FIELD_TYPE_RSV = (1 << 4)
REG_FIELD_TYPE_1C  = (1 << 5)
REG_FIELD_TYPE_1S  = (1 << 6)
REG_FIELD_TYPE_V   = (1 << 7)
REG_FIELD_TYPE_P   = (1 << 8)

__REG_FIELD_TYPES_STR__ = {REG_FIELD_TYPE_RO : "RO", REG_FIELD_TYPE_RW : "RW",
                           REG_FIELD_TYPE_WO : "WO", REG_FIELD_TYPE_RSV : "RSV",
                           REG_FIELD_TYPE_V : "V", REG_FIELD_TYPE_P : "P",
                           REG_FIELD_TYPE_1C : "1C", REG_FIELD_TYPE_1S : "1S"}
__REG_FIELD_TYPES__ = [REG_FIELD_TYPE_RO, REG_FIELD_TYPE_RW, REG_FIELD_TYPE_WO,
                       REG_FIELD_TYPE_RSV, REG_FIELD_TYPE_V, REG_FIELD_TYPE_P,
                       REG_FIELD_TYPE_1C, REG_FIELD_TYPE_1S]

RegisterLevel = lib.enums.Enum(["GREEN", "WHITE", "ORANGE", "RED", "BLACK"])
HwStates = lib.enums.Enum(["Unknown", "HW", "SAVED1", "SAVED2", "CAPTURED"])
HwTypes = lib.enums.Enum(["HFI1"], values = [1])
SupportedHWRevs = lib.enums.Enum(["A0", "B0", "B1"], values=[0x0, 0x10, 0x11])

list_intr = lambda l, r: list(set(l).intersection(r))
list_diff = lambda l, r: list(set(l).difference(r))
list_symm = lambda l, r: list(selt(l).symmetric_difference(r))

logger = initModuleLogging(__name__)

def index2hwstate_saved(idx):
    if idx == 1:
	return str(HwStates.SAVED1)
    elif idx == 2:
	return str(HwStates.SAVED2)
    else:
	raise HwError("Invalid index '%d'" % idx)

def access2str(value):
    return '/'.join([__REG_FIELD_TYPES_STR__[x] for x in __REG_FIELD_TYPES__ if \
                     value & x])

def str2access(string):
    regval = 0
    for t in string.split('/'):
        try: regval |= eval("REG_FIELD_TYPE_%s" % t.upper())
        except (SyntaxError, NameError): continue
    return regval

class Devices(object):
    def __init__(self, default_device=HwTypes.HFI1):
        self.__devlist = {}
        self.default_dev = default_device
        self.enumerateDevices()
        return

    def __len__(self):
        return len(self.__devlist)

    def __iter__(self):
        for key, dev in self.__devlist.items():
            yield (key, dev[0])

    def enumerateDevices(self):
        count = 0
        if not os.path.exists("/sys/bus/pci/devices"):
            return
        for bus_id in os.listdir("/sys/bus/pci/devices"):
            full_path = os.path.sep.join(["/sys/bus/pci/devices", bus_id])
            if os.path.isdir(full_path) and \
               os.path.exists(os.path.join(full_path, "device")):
                pciid = 0
                with open(os.path.join(full_path, "device")) as fd:
                    pciid = int(fd.readline().strip(), 16)
                for rev, ids in PciIdMap.items():
                    if pciid in ids[0]:
                        devpath = os.path.join(full_path, "infiniband")

                        unit = None
                        if os.path.exists(devpath):
                            devlist = os.listdir(devpath)
                            for dev in [x for x in devlist if \
                                        x.startswith(str(rev).lower())]:
                                break
                            try: dev, unit = dev.split('_')
                            except ValueError: continue
                            unit = int(unit)

                        unit = unit if unit is not None and \
                               unit not in self.__devlist else count

                        self.__devlist[unit] = (ids[1], pciid, bus_id, unit)
                        logger.log(INFO,
                                   "Discovered device type %s, PCI ID 0x%x, bus id %s, idx %d",
                                   rev, self.__devlist[unit][1],
                                   self.__devlist[unit][2], self.__devlist[unit][3])
                        count += 1
        return

    def createDevice(self, unit=0, *args, **kwargs):
        if unit in self.__devlist:
            dev = self.__devlist[unit]
            return dev[0](dev[1], dev[2], dev[3], *args, **kwargs)
        else:
            return None

    def createDefaultDevice(self, *args, **kwargs):
        dev_type = PciIdMap[self.default_dev]
        return dev_type[1](dev_type[0][0], None, 0, *args, **kwargs)

    def createDefaultDeviceType(self, type, *args, **kwargs):
        if type in PciIdMap:
            dev_type = PciIdMap[type]
            return dev_type[1](dev_type[0][0], None, 0, *args, **kwargs)
        return None

class HwError(exceptions.Exception): pass
class ResourceError(exceptions.Exception): pass

class Mmap(object): pass

class LockedResource(object):
    file_pattern = "/sys/kernel/debug/%s/%s_%u/%s"
    def __init__(self, device, name, unit=0, offset=0, length=0):
        self.filename = self.file_pattern % (device, device, unit, name)
        self.name = name
        self.offset = offset
        self.length = length
        self.fd = None

    def open(self):
        try:
            self.fd = open(self.filename, "r+b", 0)
        except Exception, e:
            logger.log(ERROR, "open of %s failed: %s", self.filename, e)
            pass

    def close(self):
        if self.fd:
            self.fd.close()

    def belongs(self, address):
        if address >= self.offset and address < self.offset + self.length:
            return True
        return False

    def raw_read(self, addr, length=c_ulong(8)):
        if not self.belongs(addr.value):
            raise ResourceError("%s address outside of resource range" % self.name)
        if not self.fd:
            raise ResourceError("%s resource not available" % self.name)
        address = addr.value - self.offset
        try:
            self.fd.seek(address, os.SEEK_SET)
            return self.fd.read(length.value)
        except IOError, e:
            logger.log(ERROR, "%s read failed @ %#x: %s", self.name,
                        addr.value, e)
            raise ResourceError(str(e))

    def read(self, addr, length=c_ulong(8)):
        buf = self.raw_read(addr, length)
        return struct.unpack("Q" * (length.value / 8), buf)

    def write(self, addr, value, length=c_ulong(8)):
        if not self.belongs(addr.value):
            raise ResourceError("%s address outside of resource range" % self.name)
        if not self.fd:
            raise ResourceError("%s resource not available" % self.name)
        address = addr.value - self.offset
        try:
            self.fd.seek(address, os.SEEK_SET)
            offset = 0
            while offset < length.value:
                logger.log(INFO, "%s write: %#x @ %#x", self.name, value.value, address + offset)
                self.fd.write(struct.pack("Q", value.value))
                offset += 8
            return True
        except IOError, e:
            logger.log(ERROR, "%s write failed @ %#x: %s", self.name,
                        addr.value, e)
            raise ResourceError(str(e))

class Resource(object):
    path = "/sys/bus/pci/devices/"

    def __init__(self, maps, index=0, flags=0x0):
        self.index = index
        self.flags = flags
        self.bus_id = ""
        self.device = None
        self.maps = []
        for mmap in maps:
            mm = Mmap()
            mm.offset = mmap[0]
            mm.size = mmap[1]
            mm.mm = None
            self.maps.append(mm)
        return

    def set_bus_id(self, bid):
        self.bus_id = bid
        return

    def open(self):
        logger.log(INFO, "Attempting to open %s", self.filename())
        self.device = open(self.filename(), "r+b", 0)
        if self.device:
            for mm in self.maps:
                mm.mm = mod_mmap.mmap(self.device.fileno(), mm.size,
                                      mod_mmap.MAP_SHARED,
                                      mod_mmap.PROT_READ|mod_mmap.PROT_WRITE,
                                      offset=mm.offset)

    def isopen(self):
        return isinstance(self.device, file) and \
               True in [isinstance(mm.mm, mod_mmap.mmap) for mm in self.maps]

    def close(self):
        for mmap in self.maps:
            if mmap.mm:
                mmap.mm.close()
                mmap.mm = None
        if self.device:
            self.device.close()
            self.device = None

    def filename(self):
        return os.path.join(self.path, self.bus_id, "resource%d" % self.index)

    def addr2mmap(self, addr, length=8):
        for mmap in self.maps:
            if addr >= mmap.offset and \
               (addr + length) <= (mmap.offset + mmap.size) and \
               mmap.mm and isinstance(mmap.mm, mod_mmap.mmap):
                logger.log(DEBUG, "Found mmap: 0x%x / 0x%x", mmap.offset, mmap.size);
                return mmap
        return None

    def raw_read(self, addr, length):
        if not self.isopen():
            raise ResourceError("Resource unavailable")
        if type(addr) != c_ulonglong:
            raise ResourceError("Invalid address type %s" % type(addr))
        if type(length) != c_ulong:
            raise ResourceError("Invalid length type %s" % type(length));
        if addr.value == ADDR_INVALID: return []
        if addr.value % 8:
            raise ResourceError("Address must be 8 byte aligned")
        if length.value % 8:
            raise ResourceError("Length must be multiple of 8 bytes")

        mmap = self.addr2mmap(addr.value, length.value)
        if not mmap:
            raise ResourceError("Invalid address for resource")

        logger.log(INFO, "Reading %u bytes @ 0x%x", length.value, addr.value)
        buf = ""
        logger.log(DEBUG, "Mmap offset: 0x%x", addr.value - mmap.offset)
        mmap.mm.seek(addr.value - mmap.offset)
        for idx in range(length.value / 8):
            buf += mmap.mm.read(8)
        return buf

    def read(self, addr, length=c_ulong(8)):
        try:
            buf = self.raw_read(addr, length)
            result = struct.unpack("Q" * (length.value / 8), buf)
            logger.log(INFO, "read[%s]: %dB @ %#x: %s", HwStates.HW, length.value,
                       addr.value, result)
            return result
        except IOError, e:
            logger.log(ERROR, "read[%s]: failed @ %#x: %s", HwStates.HW,
                       addr.value, e)
            raise ResourceError(str(e))

    def write(self, addr, value, length=c_ulong(8)):
        if not self.isopen():
            raise ResourceError("Resource unavailable")
        if type(addr) != c_ulonglong:
            raise ResourceError("Invalid address type %s" % type(addr))
        if type(value) != c_ulonglong:
            raise ResourceError("Invalid value type %s" % type(addr))
        if type(length) != c_ulong:
            raise ResourceError("Invalid length type %s" % type(length))
        if addr.value == ADDR_INVALID: return False
        if addr.value % 8:
            raise ResourceError("Address must be 8 byte aligned")
        if length.value % 8:
            raise ResourceError("Length must be multiple of 8 bytes")

        offset = 0
        mmap = self.addr2mmap(addr.value, length.value)
        if not mmap:
            raise ResourceError("Invalid address for resource")
        try:
            mmap.mm.seek(addr.value - mmap.offset, os.SEEK_SET)
            while offset < length.value:
                logger.log(INFO, "write: %#x @ %#x", value.value, addr.value + offset)
                mmap.mm.write(struct.pack("Q", value.value))
                offset += 8
        except IOError, e:
            logger.log(ERROR, "write: failed @ %#x: %s", addr.value + offset, e)
            raise ResourceError("Write to 0x%x failed: %s" % (addr.value + offset, e))
        return True

class RegisterFieldAttrs:
    def __init__(self):
        self.attrs = {"shift": c_ulonglong(0), "mask": c_ulonglong(0),
                      "smask": c_ulonglong(0), "access": c_ulonglong(REG_FIELD_TYPE_RW)}
    def set_field_attr(self, attr, value):
        attr = attr.lower()
        if attr in self.attrs:
            if attr == "access":
                if not isinstance(value, int) and not isinstance(value, long) and \
                   value not in __REG_FIELD_TYPES__: return False
                self.attrs[attr] = c_ulonglong(value)
                return True
            if isinstance(value, basestring):
                if value[-3:].lower() == "ull": value = value[:-3]
                if lib.diags.is_hex(value): self.attrs[attr] = c_ulonglong(int(value, 16))
                else: self.attrs[attr] = c_uint(int(value))
            elif isinstance(value, int): self.attrs[attr] = c_uint(value)
            elif isinstance(value, long): self.attrs[attr] = c_ulonglong(value)
            else: return False
            return True
        return False
    def get_field_attr(self, attr):
        attr = attr.lower()
        if attr in self.attrs: return self.attrs[attr].value
        return c_ulonglong(-1).value

class RegisterFields:
    def __init__(self):
        self.fields = {}
    def add_field(self, name):
        self.fields[name] = RegisterFieldAttrs()
    def set_field_attr(self, name, attr, value):
        if name in self.fields:
            self.fields[name].set_field_attr(attr, value)
    def get_all_str(self):
        return sorted(self.fields.keys(),
                      key = lambda x: self.fields[x].get_field_attr("shift"),
                      reverse = True)

    def get_field_attr(self, name, attr):
        return self.fields[name].get_field_attr(attr)
    def __str__(self):
        s = ""
        fields = self.get_all_str()
        if fields:
            s = "\t%-45s %-6s %-18s %-18s\n" % ("NAME", "SHIFT", "MASK", "SMASK")
            for x in fields:
                s += "\t%-45s" % x + " %-6d %#-18x %#-18x\n" % \
                     (self.fields[x].get_field_attr("shift"),
                      self.fields[x].get_field_attr("mask"),
                      self.fields[x].get_field_attr("smask"))
        return s
    def __repr__(self):
        return ",".join(["%s(%d,0x%x)" % (x, self.fields[x].get_field_attr("shift"),
                                          self.fields[x].get_field_attr("mask")) \
                         for x in self.get_all_str()])
    def __iter__(self):
        for x in self.fields.items(): yield x
    def __contains__(self, field):
        return field in self.fields

class Register:
    def __init__(self, name, address_str, address=ADDR_INVALID,
                 count=1, incr=0x8, asize=1, width=64, security=None,
                 debug=False):
        self.name = name
        self.address = address
        self.address_str = address_str
        self.width = width
        self.resetval = 0
        self.fields = RegisterFields()
        self.count = count
        self.step = incr
        self.array_size = asize
        self.is_debug = debug
        self.level = security if security else RegisterLevel.WHITE
        self.flags = 0
    def set_address(self, addr):
        self.address = addr
    def set_reset(self, value):
        self.resetval = value
    def set_flag(self, flag):
        self.flags |= flag
    def add_field(self, name, type, value):
        if name in self.fields:
            self.fields.set_field_attr(name, type, value)
        else:
            self.fields.add_field(name)
            self.fields.set_field_attr(name, type, value)
    def get_field_attr(self, name, attr):
        if name in self.fields:
            return self.fields.get_field_attr(name, attr)
    def get_fields_str(self):
        return self.fields.get_all_str()
    def is_readable(self):
        return not bool(self.flags & RF_NON_READABLE)
    def is_locked(self):
        return bool(self.flags & RF_LOCKED)
    def has_field(self, name):
        return name in self.fields
    def get_field(self, name, value):
        if name in self.fields:
            mask = self.get_field_attr(name, "mask")
            shift = self.get_field_attr(name, "shift")
            if type(value) == c_ulong: value = value.value
            return ((value >> shift) & mask)
        return 0x0
    def addressof(self, idx=0, elem=0):
        if self.address != ADDR_INVALID and \
           idx < self.count:
            base = self.address + (idx * self.step)
            if elem < self.array_size:
                return base + (elem * (self.width / 8))
        return ADDR_INVALID
    def describe(self, index=0):
        bitfields = []
        for name in self.fields.get_all_str():
            bitfields.append((name,
                              self.get_field_attr(name, "access"),
                              self.get_field_attr(name, "shift"),
                              self.get_field_attr(name, "mask"),
                              self.get_field_attr(name, "smask")))
        return (self.name, self.addressof(index), self.count,
                self.step, bitfields)
    def __str__(self):
        s = self.name + " : " + str(self.address) + "\n"
        s += str(self.fields)
        return s
    def __repr__(self):
        return "%s(%s,0x%x,%u,%u,%u,%u,0x%x,%s)[%s]" % \
               (self.__class__.__name__, self.name, self.address,
                self.width, self.count, self.step, self.array_size,
                self.resetval, self.level, repr(self.fields))

_pContainer = lambda *args: tuple(args)
pickle_proto = cPickle.HIGHEST_PROTOCOL

FieldDecode = collections.namedtuple("FieldDecode",
                                     "name width shift value")

class RegisterDecode(object):
    def __init__(self, fields=[]):
        self.fields = fields
    def __getattr__(self, name):
        if name == "fields": return self.fields
        for field in self.fields:
            if name == field.name: return field
        raise AttributeError("RegisterDecode does contain field")
    def __iter__(self):
        for field in self.fields:
            yield field

# Python's print is a builtin function and therefore, we can't
# get a pointer to it. So, we have to have a wrapper.
def default_print(fmt, *args): print fmt % args

class Hw(object):
    hw_type = None
    dbname = None
    resource = None

    def __init__(self, pid, bid, unit=0, datapath=None, tests=False):
        self.globals = {}
        self.registers = []
        self.pciregisters = []
        self.pci_id = pid
        self.bus_id = bid
        self.pci_device = bid
        self.unit = unit
        self.datapath = datapath or os.path.dirname(os.path.dirname(__file__))
        self.register_index_names = {}
        self.memmap = []
        self.memstate = [None, None]
        self.hwstate = None
        self.memstate_idx = 0
        self.active_state = HwStates.HW
        self.datahash = None
        self.hashset = []
        self.test_mode = tests
        self.implementations =  ["RTL silicon",	"RTL VCS simulation",
                                 "RTL FPGA emulation", "Functional simulator",
                                 "Unknown"]
        libname = util.find_library("c")
        self.libc = CDLL(libname)
        self.printf = default_print
        self.locked = []
        if self.resource:
            self.resource.set_bus_id(self.bus_id)


    def get_hw(self):
        return self.hw_type

    def __del__(self):
        self.close()

    def __find_state_segment(self, state, addr):
        if isinstance(state, tuple): state = state[2]
        for segment in state:
            if addr >= segment[0] and addr < (segment[0] + segment[1]):
                return segment
        return None

    def __state_translate(self, state):
        if isinstance(state, str):
            state = state.upper()
            if state == str(HwStates.HW): return HwStates.HW
	    elif state == str(HwStates.SAVED1): return HwStates.SAVED1
            elif state == str(HwStates.SAVED2): return HwStates.SAVED2
            elif state == str(HwStates.CAPTURED): return HwStates.CAPTURED
            raise HwError("Unknown state '%s'" % state)
        elif isinstance(state, lib.enums.EnumType):
            if state not in HwStates:
                raise HwError("Unknown state '%s'" % state)
            return state

    def open(self):
        # Check whether the device is enabled. If not, enable it.
        # This can happen if the driver has never been loaded. In that
        # case, the kernel never enables the device.
        reg = lib.pci.read_register(self.pci_device, 0x4)
        if reg & 0x3 == 0:
            lib.pci.enable_device(self.pci_device)

        if not self.resource:
            return False
        try:
            self.resource.open()
        except Exception, e:
            raise HwError(e)
        for l in self.locked:
            try:
                l.open()
            except Exception, e:
                raise HwError(e)
        return True

    def close(self):
        if self.resource:
            self.resource.close()
            self.resource = None
        if self.locked:
            for l in self.locked:
                l.close()
            del self.locked[:]
            self.locked = []
        self.unit = -1
        return

    def generate_mem_map(self, register_set=None):
        __memmap = []
        if register_set is None: register_set = self.registers
        if not isinstance(register_set, list): return False
        for elem in register_set:
            if not isinstance(elem, Register): return False
        if register_set:
            memsize, last = 0, None
            for reg in register_set:
                if reg.is_readable():
                    if reg.address > memsize:
                        memsize = reg.address
                        last = reg
            memsize += (last.step * last.count) + \
                       ((last.width / 8) * last.array_size)
            memmap = [0] * memsize
            for reg in register_set:
                if reg.is_readable() and reg.address != ADDR_INVALID:
                    size = reg.width / 8
                    for ridx in range(reg.count):
                        for aidx in range(reg.array_size):
                            addr = self.addressof(reg, ridx, aidx)
                            memmap[addr:addr+size] = [1] * size
            prev = 0
            for idx in range(len(memmap)):
                if memmap[idx] == 1 and prev == 0: start = idx
                elif memmap[idx] == 0 and prev == 1:
                    __memmap.append((start, idx-start))
                elif memmap[idx] == 1 and idx == len(memmap) - 1:
                    __memmap.append((start, idx-start+1))
                prev = memmap[idx]
        return sorted(__memmap, key=lambda x: x[0])

    def md5(self, globs=None, regs=None, conf=None, mmap=None):
        if globs == None: globs = self.globals
        if regs == None: regs = self.registers
        if conf == None: conf = self.pciregisters
        if mmap == None: mmap = self.memmap
        regs = sorted(regs, key=lambda x: x.name)
        conf = sorted(conf, key=lambda x: x.name)
        h = hashlib.md5()
        h.update(repr(sorted(globs)))
        h.update(repr(regs))
        h.update(repr(conf))
        h.update(repr(mmap))
        return h.hexdigest()

    def load_hw_data(self, datadir="data", version=None):
        datadir = os.sep.join([self.datapath, datadir])
        datadir = os.path.realpath(datadir)
        dirlist = os.listdir(datadir)
        matching_files = {}
        if version == None:
            version = self.hw_revision_raw()[0] & 0x3f
            try: version = SupportedHWRevs.byval(version)
            except AttributeError, e:
                return DB_ERR_VER
        elif not isinstance(version, lib.enums.EnumType):
            return DB_ERR_VER
        logger.log(DEBUG, "Using HWRev: 0x%x", version)
        loaded = False
        ret = DB_ERR_NO_DB_LOADED
        for filename in dirlist:
            if filename.startswith(self.dbname) and \
               filename.endswith(".r"):
                filename = os.sep.join([datadir, filename])
                logger.log(DEBUG, "Loading database file %s", filename)
                # determine the type of the data file
                format = DbFileFormats.FLAT
                try:
                    fd = open(filename, 'r')
                    header = fd.read(2)
                    if header == "BZ": format = DbFileFormats.BZIP
                    fd.close()
                except IOError, e:
                    logger.log(ERROR, "Failed to open DB file %s: %s",
                               filename, e)
                    continue

                logger.log(DEBUG, "File format: %s", format)
                if format == DbFileFormats.BZIP:
                    fd = bz2.BZ2File(filename, "rb")
                else: fd = open(filename, 'rb')

                try:
                    __hash, __ver, __globs, __registers, __confregs, \
                            __memmap, __reg_idxs = cPickle.load(fd)
                except (cPickle.UnpicklingError, AttributeError,
                        EOFError, ImportError, IndexError), e:
                    logger.log(ERROR, "DB file %s did not unpack correctly: %s",
                               os.path.basename(filename), e)
                    ret = DB_ERR_UNPACK
                    continue
                except ValueError:
                    logger.log(INFO,
                               "DB file %s does not match format. Skipping...",
                               os.path.basename(filename))
                    ret = DB_ERR_FORMAT
                    continue
                finally: fd.close()
                # Is the database version valid and does this database
                # match the HW version we are running on
                if version == None:
                    try: version = SupportedHWRevs.byval(__ver)
                    except AttributeError:
                        ret = DB_ERR_HWREV_UNSUP
                        continue
                elif (__ver & 0xF0) != (version.get() & 0xF0):
                    logger.log(WARN,
                               "DB file rev for %s does not match HW " + \
                               "rev 0x%x != 0x%x", os.path.basename(filename),
                               __ver, version.get())
                    ret = DB_ERR_HWREV
                    continue
                matching_files[filename] = (__hash, __ver, __globs, __registers,
                                            __confregs, __memmap, __reg_idxs)

        # Create a sorted list of the usable files that are not newer than
        # what version holds, the files then are sorted based on stepping to
        # make sure that the values found in stepping specific files prevail
        file_versions = []
        for matching_file, values in matching_files.items():
            if values[1] > version.get():
                continue
            file_versions.append((matching_file, values[1]))

        sorted_files = sorted(file_versions, key=lambda x: x[1], reverse=True)
        usable_files = [x for x in sorted_files if x[1] == sorted_files[0][1]]

        for filename, stepping_version in usable_files:
            logger.log(INFO, "Using database file: %s", filename)
            __hash, __ver, __globs, __registers, __confregs, __memmap, \
		    __reg_idxs = matching_files[filename]

            # Now, we update...
            if __hash not in self.hashset: self.hashset.append(__hash)
            loaded = True
            # Update the dictionary of globals
            __intr = list_intr(__globs.keys(), self.globals.keys())
            map(__globs.__delitem__, __intr)
            self.globals.update(__globs)

            # Add any new registers
            for reg in __registers: self.add_register(reg)

            # Add any new config registers
            for reg in __confregs: self.add_config_register(reg)

            # Update the memory map
            if not self.memmap: self.memmap = __memmap
            else:
                __current = self.memmap[:]
                self.memmap = []
                for segment in __memmap:
                    if segment[0] in [s[0] for s in __current]:
                        # The segment already exists, find it's index
                        for idx in xrange(len(__current)):
                            if segment[0] == __current[idx][0]: break
                        # Check if the new segment is a superset of the
                        # old. If so, use the new segment.
                        if segment[1] > __current[idx][1]:
                            self.memmap.append(segment)
                        del __current[idx]
                    else:
                        self.memmap.append(segment)
                # Add everything that is left over
                self.memmap += __current

            # Update symbolic index names
            __intr = list_intr(__reg_idxs.keys(),
                               self.register_index_names.keys())
            map(__reg_idxs.__delitem__, __intr)
            self.register_index_names.update(__reg_idxs)

        # If we haven't loaded anything, don't bother with the rest
        if not loaded: return ret

        # Sort the newly generate map
        self.memmap = sorted(self.memmap, key=lambda x: x[0])
        # Concatinate adjacent segments
        to_remove = []
        for idx in range(1, len(self.memmap)):
            prev, this = self.memmap[idx-1], self.memmap[idx]
            if this[0] == prev[0] + prev[1]:
                self.memmap[idx-1] = (prev[0], prev[1] + this[1])
                to_remove.append(idx)
        for idx in to_remove:
            del self.memmap[idx]
        self.datahash = self.md5()
        logger.log(DEBUG, "Computed hash on loaded data: %s", self.datahash)
        self.hashset.append(self.datahash)
        return 0

    def set_print_func(self, func):
        if func and callable(func): self.printf = func
        return

    def addressof(self, name, index=0, elem=0):
        base = 0L
        if not isinstance(name, Register):
            reg = self.get_register_by_name(name)
            if not reg:
                if name in self.globals:
                    return self.globals[name].value
                raise HwError("Unknown register %s" % name)
        else: reg = name

        if index >= reg.count:
            raise HwError("Register %s has only %d instances" % \
                            (name, reg.count))
        addr = reg.addressof(index, elem)
        if addr == ADDR_INVALID and reg.address_str and \
           reg.address_str[0] == '(':
            addrstr = reg.address_str[1:-1]
            addrstr = [x.strip() for x in addrstr.split("+")]
            for part in addrstr:
                if lib.diags.is_hex(part): base += c_ulonglong(int(part, 16)).value
                else: base += \
                      c_ulonglong(self.addressof(part, 0, 0)).value
            addr = reg.addressof(index, elem)
        return addr

    def __add_register(self, reg, array):
        if isinstance(reg, Register):
            if not filter(lambda x: x.name == reg.name, array):
                array.append(reg)
                return True
        return False

    def add_register(self, reg):
        return self.__add_register(reg, self.registers)

    def add_config_register(self, reg):
        return self.__add_register(reg, self.pciregisters)

    def add_global(self, name, value):
        if value[-1] == "L": value = value[:-1]
        if name not in self.globals or \
           self.globals[name].value == 0:
            if lib.diags.is_hex(value):
                self.globals[name] = c_ulonglong(int(value, 16))
            else: self.globals[name] = c_uint(int(value))
        return

    def add_register_index(self, name, value):
        if name not in self.register_index_names:
            self.register_index_names[name] = value

    def get_register_by_name(self, name):
        for reg in self.registers + self.pciregisters:
            if name == reg.name: return reg
        return None

    def get_register_by_addr(self, addr, info=None):
        for reg in self.registers:
            for idx in range(reg.count):
                regaddr = reg.address + (idx * reg.step)
                if addr == regaddr or \
                   (addr > regaddr and \
                    addr < (regaddr + (reg.array_size *
                                       (reg.width / 8)))):
                    r = deepcopy(reg)
                    r.address = addr
                    r.count = 1
                    if reg.array_size: r.array_size = 1
                    if info is not None and isinstance(info, list):
                        info.append(idx)
                        info.append((addr - regaddr) / (reg.width / 8))
                    return r
        return None

    def get_register(self, reg):
        if isinstance(reg, Register): return reg
        elif isinstance(reg, int) or isinstance(reg, long):
            return self.get_register_by_addr(reg)
        elif isinstance(reg, str):
            return self.get_register_by_name(reg)
        else: return None

    def get_registers(self):
        return self.registers

    def get_config_registers(self):
        return self.pciregisters

    def get_register_names(self):
        return [reg.name for reg in self.registers]

    def get_config_register_names(self):
        return [reg.name for reg in self.pciregisters]

    def get_global_names(self):
        return sorted(self.globals.keys())

    def get_global_by_name(self, name):
        return self.globals.get(name, None)

    def get_globals(self):
        return sorted(self.globals.items(), key=lambda x: x[0])

    def get_global(self, name):
        gl = self.get_global_by_name(name)
        if gl: return gl.value
        return None

    def get_register_fields(self, regname):
        register = self.get_register(regname)
        if register:
            return register.get_fields_str()
        return []

    def get_register_count(self, regname):
        reg = self.get_register(regname)
        if reg: return reg.count
        return 0

    def get_register_array_size(self, regname):
        reg = self.get_register(regname)
        if reg: return reg.array_size
        return 0

    def get_register_index_names(self):
        return self.register_index_names.keys()

    def get_register_index(self, name):
        return self.register_index_names.get(name, 0)

    def has_secure(self):
        return set([reg.level.get() for reg in self.registers]) != \
               set([RegisterLevel.GREEN.get()])

    def decode_register(self, regname, value, filter=[]):
        fieldvals = []
        reg = self.get_register(regname)
        fields = reg.get_fields_str()
        for field in fields:
            if filter and field not in filter: continue
            mask = reg.get_field_attr(field, "mask")
            shift = reg.get_field_attr(field, "shift")
            bits = 0
            cmask = mask
            while cmask:
                bits += 1
                cmask >>= 1
            val = ((value & (mask << shift)) >> shift)
            fieldvals.append(FieldDecode(field, bits, shift, val))
        return RegisterDecode(fieldvals)

    def __findLockedResource(self, addr):
        for l in self.locked:
            if l.belongs(addr.value):
                return l
        return None

    def read(self, addr, length=c_ulong(8), use_state=None):
        reg = self.get_register_by_addr(addr.value)
        if reg and not reg.is_readable():
            return (0,)

        if not use_state: use_state = self.active_state
        else: use_state = self.__state_translate(use_state)
        if use_state == HwStates.HW:
            if reg and reg.is_locked():
                lockedResource = self.__findLockedResource(addr)
                if lockedResource:
                    try:
                        result = lockedResource.read(addr, length)
                    except ResourceError, e:
                        raise HwError(e)
                else:
                    raise HwError("Locked CSR does not belong to any resource")
            else:
                try:
                    result = self.resource.read(addr, length)
                except ResourceError, e:
                    raise HwError(e)
        else:
            state = self.get_state(use_state)
            if not state: raise HwError("Missing or wrong saved state")
            segment = self.__find_state_segment(state, addr.value)
            if not segment:
                raise HwError("Could not find register in state %s" % \
                              use_state)
            off = addr.value - segment[0]
            size = length.value / 8
            format = "Q" * size
            logger.log(INFO, "read[%s]: %dB @ %#x", use_state,
                       length.value, addr.value)
            result = list(struct.unpack(format,
                                        segment[2][off:off+length.value]))
            logger.log(INFO, "read[%s]: result %#x", use_state, result[-1])
        return result

    def write(self, addr, value, length=c_ulong(8)):
        if self.active_state != HwStates.HW:
            raise HwError("Can not write when using saved states")
        reg = self.get_register_by_addr(addr.value)
        if reg and reg.is_locked():
            lockedResource = self.__findLockedResource(addr)
            if lockedResource:
                try:
                    return lockedResource.write(addr, value, length)
                except ResourceError, e:
                    raise HwError(e)
            else:
                raise HwError("Locked CSR does not belong to any resource")
        else:
            try:
                return self.resource.write(addr, value, length)
            except ResourceError, e:
                raise HwError(e)

    def read_register(self, regname, index=0, elem=0, state=None):
        if not self.get_register(regname):
            raise HwError("Unknown register %s" % regname)
        addr = self.addressof(regname, index, elem)
        return self.read(c_ulonglong(addr), use_state=state)[0]

    def write_register(self, regname, index, elem, value):
        if not self.get_register(regname):
            raise HwError("Unknown register %s" % regname)
        addr = self.addressof(regname, index, elem)
        if isinstance(value, int) or isinstance(value, long):
            value = c_ulonglong(value)
        return self.write(c_ulonglong(addr), value)

    def __register_field_access_mask(self, register, value):
        warning_shown = False
        for f in register.get_fields_str():
            access = register.get_field_attr(f, "access")
            if access & REG_FIELD_TYPE_RO or access & REG_FIELD_TYPE_RSV or \
                   (access & (REG_FIELD_TYPE_RW | REG_FIELD_TYPE_1C) ==
                    (REG_FIELD_TYPE_RW | REG_FIELD_TYPE_1C)):
                value &= ~register.get_field_attr(f, "smask")
            elif access & REG_FIELD_TYPE_WO and not warning_shown:
                self.printf("Writing to CSRs which contian Write-Only bit " + \
                            "fields could have unwanted side effects due to " + \
                            "read/modify/write operation!!")
                warning_shown = True
        return value

    def write_register_field(self, regname, index, elem, field, value):
        register = self.get_register(regname)
        if not register:
            raise HwError("Unknown register %s" % regname)
        if not register.has_field(field):
            raise HwError("Register %s does not have bitfield name %s" % \
                          (regname.name, field))

        addr = self.addressof(register, index, elem)
        regvalue = self.read(c_ulonglong(addr))[0]
        regvalue = self.__register_field_access_mask(register, regvalue)

        shift = register.get_field_attr(field, "shift")
        mask = register.get_field_attr(field, "mask")
        smask = register.get_field_attr(field, "smask")

        regvalue &= ~smask
        regvalue |= ((value.value & mask) << shift)
        return self.write(c_ulonglong(addr), c_ulonglong(regvalue))

    def read_config_register(self, regname, index=0, elem=0):
        if not self.pci_device:
            raise HwError("PCI device not known")
        if index or elem:
            raise HwError("PCI registers do not supported indexing")
        pcireg = self.get_register(regname)
        if not pcireg:
            raise HwError("Failed to find PCI Configuration register")
        value = lib.pci.read_register(self.pci_device, pcireg.addressof())
        if value is False:
            raise HwError("Failed to read PCI configuration register")
        return value

    def write_config_register(self, regname, index, elem, value):
        if not self.pci_device:
            raise HwError("PCI device not known")
        if index or elem:
            raise HwError("PCI registers do not support indexing")
        pcireg = self.get_register(regname)
        if not pcireg:
            raise HwError("Failed to find PCI Configuration register")
        if isinstance(value, c_ulonglong): value = value.value
        return lib.pci.write_register(self.pci_device, pcireg.addressof(), value)

    def write_config_register_field(self, regname, index, elem, field, value):
        reg = self.get_register(regname)
        _v = self.read_config_register(regname, index, elem)
        _v = self.__register_field_access_mask(reg, _v)

        shift = reg.get_field_attr(field, "shift")
        mask = reg.get_field_attr(field, "mask")
        smask = reg.get_field_attr(field, "smask")

        _v &= ~smask
        _v |= ((value.value & mask) << shift)
        return self.write_config_register(regname, index, elem, c_ulonglong(_v))

    def hw_revision_raw(self):
        return tuple(self.read(c_ulonglong(0), c_ulong(16)))

    def hw_revision(self):
        rev1_reg = self.get_register_by_name("CceRevision")
        rev2_reg = self.get_register_by_name("CceRevision2")
        ccestatus = self.get_register_by_name("CceStatus")

        if not rev1_reg or not rev2_reg:
            return {"unit": self.unit, "impl": 0, "impl_str": "unknown",
                    "impl_rev": 0, "chip_rev_major": 0, "chip_rev_minor": 0,
                    "status": {}}

        rev1_val = self.read(c_ulonglong(rev1_reg.address))[0]
        rev2_val = self.read(c_ulonglong(rev2_reg.address))[0]

        chipmajor = rev1_reg.get_field("ChipRevMajor", rev1_val)
        chipminor = rev1_reg.get_field("ChipRevMinor", rev1_val)

        impl = rev2_reg.get_field("ImplCode", rev2_val)
        implrev = rev2_reg.get_field("ImplRevision", rev2_val)

        if impl >= len(self.implementations):
            impl = len(self.implementations) - 1
        unit_info = {"unit": self.unit,
                     "impl": impl,
                     "impl_str": self.implementations[impl],
                     "impl_rev": implrev,
                     "chip_rev_major": chipmajor,
                     "chip_rev_minor": chipminor
                     }
        cceinfo = {}
        if ccestatus:
            val = self.read(c_ulonglong(ccestatus.address))[0]
            bits = self.decode_register(ccestatus, val)
            for bit in bits:
                if bit[3]: cceinfo[bit[0]] = True
        unit_info["status"] = cceinfo
        return unit_info

    def capture_state(self):
        state = []
        for addr, size in self.memmap:
            end = addr + size
            found = False
            # if contained in a lock range, use the lock read function
            for l in self.locked:
                l_start = l.offset
                l_end = l.offset + l.length
                # expect the map to be fully contained in a lock range
                if l_start <= addr < l_end and l_start < end <= l_end:
                    logger.log(INFO, "capture: %s read %dB @ %#x", l.name, size, addr)
                    try:
                        readin = l.raw_read(c_ulonglong(addr), c_ulonglong(size))
                        state.append((addr, size, readin))
                    except ResourceError, e:
                        logger.log(ERROR, "capture: %s failed @ %#x: %s", l.name, addr, e)
                        # assume that it is OK if the lock data is not available
                    found = True
                    break
                # sanity check: the map must be fully contained
                if l_start <= addr < l_end or l_start < end <= l_end:
                    raise HWError("capture: map range (0x%x, 0x%x) not fully contained in a lock range (0x%x, 0x%x)" % (addr, end, l_start, l_end))
            if found:
                continue
            # otherwise, use the standard access
            try:
                logger.log(INFO, "capture: read %dB @ %#x", size, addr)
                readin = self.resource.raw_read(c_ulonglong(addr),
                                                c_ulonglong(size))
                state.append((addr, size, readin))
            except ResourceError, e:
                logger.log(ERROR, "capture: failed @ %#x: %s", addr, e)
                raise HwError(str(e))
        self.hwstate = (time.time(), None, state)
        return True

    def have_saved_state(self, state=None):
        if state:
            state = self.__state_translate(state)
            if state == HwStates.CAPTURED and self.hwstate: return True
            else:
                idx = state - HwStates.SAVED1
                if self.memstate[idx]: return True
        if self.memstate[0] or self.memstate[1]: return True
        return False

    def get_state(self, state=HwStates.SAVED1):
        if state == HwStates.CAPTURED and self.hwstate: return self.hwstate
        state -= 2
        if len(self.memstate) >= state and self.memstate[state]:
            return self.memstate[state]
        return None

    def switch_state(self, state=str(HwStates.HW)):
        if not state: raise HwError("Unspecified HW state")
        self.active_state = self.__state_translate(state)
        return self.active_state

    def save_state(self, filename, state=None):
        if not state or state == HwStates.Unknown:
            state = (self.memstate_idx - 1) % len(self.memstate)
            if not self.memstate[state]:
                raise HwError("No matching stored HW state to save")
            statedata = self.memstate[state]
        else:
            state = self.__state_translate(state)
            if state == HwStates.HW and not self.hwstate:
                raise HwError("Raw HW state can not be saved. Use captured state")
            else: state = HwStates.CAPTURED
            if state == HwStates.CAPTURED:
                if not self.hwstate:
                    raise HwError("HW state needs to be captured before saving")
                statedata = self.hwstate
            else:
                idx = state - HwStates.SAVED1
                if not self.memstate[idx]: return False
                statedata = self.memstate[idx]
        filename = os.path.abspath(os.path.expanduser(filename))
        dirname = os.path.dirname(filename)
        if os.path.isdir(dirname) and os.access(dirname, os.W_OK):
            dumpstate = []
            for segment in statedata[2]:
                mem_str = StringIO.StringIO(string_at(segment[2], segment[1]))
                dumpstate.append((segment[0], segment[1], mem_str))
            try:
                fd = open(filename, 'wb')
                cPickle.dump(_pContainer(self.datahash, statedata[0], filename,
                                         dumpstate),
                             fd, protocol=pickle_proto)
                fd.close()
                return True
            except IOError: pass
        return False

    def load_state(self, filename):
        filename = os.path.abspath(os.path.expanduser(filename))
        if os.path.exists(filename) and os.access(filename, os.R_OK):
            try:
                fd = open(filename, 'rb')
                data = cPickle.load(fd)
                if not isinstance(data, tuple): data = ("", 0, None, data)
                if not self.test_mode and data[0] not in self.hashset:
                    return False
                memory = []
                for segment in data[3]:
                    cdata = create_string_buffer(segment[1])
                    cdata.raw = segment[2].getvalue()
                    segment[2].close()
                    memory.append((segment[0], segment[1], cdata))
                idx = self.memstate_idx
                self.memstate_idx = (self.memstate_idx + 1) % \
                                    len(self.memstate)
                self.memstate[idx] = (data[1], filename, memory)
                fd.close()
                return HwStates.byname(index2hwstate_saved(idx+1))
            except IOError: pass
        return False

    def describe_states(self):
        states = []
        if self.hwstate: states.append((str(HwStates.CAPTURED), ) + self.hwstate[:2])
        for idx, state in enumerate(self.memstate):
            if state:
                states.append((index2hwstate_saved(idx+1),) + state[:2])
        return states

    def diff_states(self, state1, state2, callback=None):
        if not self.have_saved_state(state1):
            raise HwError("Missing state %d" % state1)
        if not self.have_saved_state(state2):
            raise HwError("Missing state %d" % state2)

        # Translate the states. If we are given bad states,
        # the translation function will raise an exception
        numstate1 = self.__state_translate(state1)
        numstate2 = self.__state_translate(state2)

        hwstate1 = self.get_state(numstate1)
        hwstate2 = self.get_state(numstate2)

        if isinstance(hwstate1, tuple): hwstate1 = hwstate1[2]
        if isinstance(hwstate2, tuple): hwstate2 = hwstate2[2]

        diffs = []

        for segidx in range(len(hwstate1)):
            # If the two state segments are identical, there is no
            # need to look at individual registers.
            if self.libc.memcmp(hwstate1[segidx][2],
                                hwstate2[segidx][2],
                                hwstate1[segidx][1]):
                # There is a difference between the two memory
                # segments. This means we have to start looking at
                # the individual registers within the segments.
                baseaddr, size, _ = hwstate1[segidx]
                addr = baseaddr
                while addr < baseaddr + size:
                    indexes = []
                    reg = self.get_register_by_addr(addr, indexes)
                    val1 = self.read(c_ulonglong(addr),
                                     c_ulonglong(reg.width/8), state1)
                    val2 = self.read(c_ulonglong(addr),
                                     c_ulonglong(reg.width/8), state2)
                    if val1 != val2:
                        if callback and callable(callback):
                            callback(reg, indexes[0], indexes[1],
                                     val1[0], val2[0])
                        else:
                            diffs.append((reg, indexes[0], indexes[1],
                                          val1[0], val2[0]))
                    addr += reg.width/8
        return diffs or True

class HFI1(Hw):
    hw_type = HwTypes.HFI1
    dbname = "hw_data."
    devname  = "hfi1_"
    resource = Resource([(0, 0x1200000),
                         (0x1200000, 0x80000),
                         (0x1300000, 0xd00000),
                         (0x2000000, 0x800000)], 0)

    def __init__(self, pid, bid, unit=0, datapath=None, tests=False):
        super(HFI1, self).__init__(pid, bid, unit, datapath, tests)
        self.locked = [LockedResource(self.devname[:-1], "lcb", unit, 0x00601000, 0x1000)]

    def open(self):
        driver_loaded = False
        result = super(HFI1, self).open()

        if result:
            dirlist = os.listdir(DEV_DIR_PATH)
            for filename in dirlist:
                if filename.startswith(self.devname):
                    driver_loaded = True
                    break
            if not driver_loaded:
                raise HwError("HFI Driver needs to be loaded.")

        return result

PciIdMap = {
    HwTypes.HFI1 : ((0x24f0, 0x24f1), HFI1),
}
