# INTEL CONFIDENTIAL
# Copyright 2014 2015 Intel Corporation All Rights Reserved.
#
# The source code contained or described herein and all
# documents related to the source code ("Material") are
# owned by Intel Corporation or its suppliers or licensors.
# Title to the Material remains with Intel Corporation or
# its suppliers and licensors. The Material may contain trade
# secrets and proprietary and confidential information of Intel
# Corporation and its suppliers and licensors, and is protected
# by worldwide copyright and trade secret laws and treaty
# provisions. No part of the Material may be used, copied,
# reproduced, modified, published, uploaded, posted, transmitted,
# distributed, or disclosed in any way without Intel's prior express
# written permission.
#
# No license under any patent, copyright, trade secret or other
# intellectual property right is granted to or conferred upon you
# by disclosure or delivery of the Materials, either expressly, by
# implication, inducement, estoppel or otherwise. Any license under
# such intellectual property rights must be express and approved by
# Intel in writing.
#
# Unless otherwise agreed by Intel in writing, you may not remove or
# alter this notice or any other notice embedded in Materials by Intel
# or Intel's suppliers or licensors in any way.
from ctypes import *
import lib.diags as libdiags
import exceptions
import sys

STRUCT_ENDIAN_NATIVE = 0
STRUCT_ENDIAN_BIG    = 1
STRUCT_ENDIAN_LITTLE = 2

# Default byte order for bitstream conversion is based on
# the native byte order
__global_struct_list = []
__bits_per_word = 16
__struct_elem_types = {
    0 : "units",
    2 : "word",
    4 : "dword",
    8 : "qword"
    }

class StructError(exceptions.Exception): pass

STRUCT_TEMPLATE = "class %s(StructBase,%s):\n\t_pack_ = 1\n" + \
                  "\t_fields_ = %s\n\torder = %d\n\t_native = %s"

class StructBase(object):
    """The StructBase class is inherited by all defined
    structure classe and provides enhanced functionality needed
    by the hfidiag tool."""
    bytes = 0
    order = STRUCT_ENDIAN_NATIVE
    def get_fields(self, name=None):
        if name and name in self.get_field_names():
            return getattr(self, name)
        return [getattr(self, name) for name in self.get_field_names()]

    def get_field_names(self):
        return [x[0] for x in self._fields_]

    def get_field_by_name(self, name):
        if name in self.get_field_names():
            return getattr(self, name)
        return None

    def describe(self):
        flds = []
        for f in self._fields_:
            attr = getattr(type(self), f[0])
            flds.append((f[0], (8 * attr.offset) + (attr.size & 0xFFFF),
                         attr.size >> 16))
        return flds
    def is_le(self):
        if self.order == STRUCT_ENDIAN_LITTLE or \
           (self.order == STRUCT_ENDIAN_NATIVE and \
            sys.byteorder == 'little'): return True
        return False
    def is_be(self):
        return not self.is_le()

def define_struct(name, fields, order=STRUCT_ENDIAN_NATIVE, wordsize=2):
    """Define a structure class based on bitfields in the 'fields'
    argument. This allows for the defintion of arbitrary header
    and/or structures structures.
    The newly defined structure class is inserted into the modules
    global dictionary and becomes part of the module. Once defined,
    header and/or structures can be instanciated like any other
    class."""
    __fieldsize,__fields_spec, __bits = \
                               "c_uint%d" % (__bits_per_word * wordsize), [], 0
    for field in fields:
        if not isinstance(field, tuple) or \
           not isinstance(field[0], basestring) or \
           not isinstance(field[1], int):
            raise StructError("Structure fields must be (<name>, <bitwidth>) tuples")
        __fields_spec.append("(" + ",".join(['"'+str(field[0])+'"',
                                             __fieldsize,
                                             str(field[1])]) + ")")
        __bits += field[1]
    __fields_spec = "[%s]" % ",".join(__fields_spec)
    if order == STRUCT_ENDIAN_BIG:
        native_type = ".".join([__fieldsize, "__ctype_be__"])
        struct = "BigEndianStructure"
    elif order == STRUCT_ENDIAN_LITTLE:
        native_type = ".".join([__fieldsize, "__ctype_le__"])
        struct = "LittleEndianStructure"
    elif order == STRUCT_ENDIAN_NATIVE:
        native_type = __fieldsize
        struct = "Structure"
    else: raise StructError("Unsupported byte order")
    class_code = STRUCT_TEMPLATE % (name, struct, __fields_spec,
                                    order, native_type)
    try:
        exec(compile(class_code, "<string>", "exec"), globals())
        __global_struct_list.append(name)
    except Exception, e: raise StructError(e)
    return True

def struct_elems(struct):
    """Return the number of structure 'elements'"""
    return sizeof(struct) / sizeof(struct._native)

def struct_elem_size(struct):
    """Return the size of a single structure 'element'"""
    return sizeof(struct._native)

def struct_elem_name(struct):
    """Return word name for structure 'element' type"""
    size = struct_elem_size(struct)
    if size not in __struct_elem_types: size = 0
    return __struct_elem_types[size]

def defined_structs():
    """Return a list with the names of all defined structuress."""
    return __global_struct_list

def get_struct_by_name(name):
    """Return an instance of the named structure. The new instance is
    empty - all bitfields have a value of '0'."""
    if name in __global_struct_list and name in globals():
        klass = globals()[name]
    else: klass = None
    return klass and klass() or klass

def struct_set_from_instance(struct, bytestream):
    """Sets all bitfields of the input structure to values from 'bytestream'.
    The input bitstream is properly converted to match the byte order of the
    structure. The function returns a new structure instance with all
    bitfields set."""
    length = len(bytestream[2:]) if libdiags.is_hex(bytestream) else \
             len(bytestream)
    if length != sizeof(struct) * 2:
        raise StructError("Byte stream is not long enough for structure")
    bytestream = str2elems(struct, bytestream, True)
    # bytestream is interpreted in network order (Big Endian) so we have to
    # reverse the order of the DWords but don't byteswap them
    if struct.order == STRUCT_ENDIAN_NATIVE: bytestream.reverse()
    bytes = (struct._native * (sizeof(struct) / sizeof(struct._native)))(*bytestream)
    return cast(pointer(bytes), POINTER(type(struct))).contents

def struct_set_from_name(name, bytestream):
    """Return a new instance of the named structure with bitfields set to
    value from 'bytestream'. The input bitstream is properly byteswapped
    to match the byte order of the struct."""
    if name in globals(): klass = globals()[name]
    else: raise StructError("Structure %s does not exist" % name)
    struct = klass()
    return struct_set_from_instance(struct, bytestream)

def __str2size(string, size=2, asint=False):
    if libdiags.is_hex(string): string = string[2:]
    string = string if len(string) % size == 0 else \
             "0" * (size - (len(string) % size)) + string
    bytes = [string[i:i+size] for i in range(0, len(string), size)]
    if asint: bytes = [int(x, 16) for x in bytes]
    return bytes

def str2elems(struct, string, asint=False):
    """Split a hex string representation of a bytestream into an array of
    structure 'elements'."""
    return __str2size(string, sizeof(struct._native) * 2, asint)

def str2bytes(string, asint=False):
    """Split a hex string representation of a bitstream into an array of
    bytes."""
    return __str2size(string, 2, asint)

def str2words(string, asint=False):
    """Split a hex string representation of a bitstream into an array of
    words (2 bytes)."""
    return __str2size(string, 4, asint)

def str2dwords(string, asint=False):
    """Split a hex string representation of a bitstream into an array of
    double words (4 bytes)."""
    return __str2size(string, 8, asint)

def str2qwords(string, asint=False):
    """Split a hex string representation of a bitstream into an array of
    quad words (8 bytes)."""
    return __str2size(string, 16, asint)

