# INTEL CONFIDENTIAL
# Copyright 2014-2017 Intel Corporation All Rights Reserved.
#
# The source code contained or described herein and all
# documents related to the source code ("Material") are
# owned by Intel Corporation or its suppliers or licensors.
# Title to the Material remains with Intel Corporation or
# its suppliers and licensors. The Material may contain trade
# secrets and proprietary and confidential information of Intel
# Corporation and its suppliers and licensors, and is protected
# by worldwide copyright and trade secret laws and treaty
# provisions. No part of the Material may be used, copied,
# reproduced, modified, published, uploaded, posted, transmitted,
# distributed, or disclosed in any way without Intel's prior express
# written permission.
#
# No license under any patent, copyright, trade secret or other
# intellectual property right is granted to or conferred upon you
# by disclosure or delivery of the Materials, either expressly, by
# implication, inducement, estoppel or otherwise. Any license under
# such intellectual property rights must be express and approved by
# Intel in writing.
#
# Unless otherwise agreed by Intel in writing, you may not remove or
# alter this notice or any other notice embedded in Materials by Intel
# or Intel's suppliers or licensors in any way.
import os
import sys
import imp
import signal
import fcntl
import struct
import termios
import inspect
from pydoc import pipepager

##############################################################
##                                                          ##
##  !! IMPORTANT !! IMPORTANT !! IMPORTANT !! IMPORTANT !!  ##
##                                                          ##
## Use our own readline module because ours is compiled     ##
## against libedit rather than libreadline. This is         ##
## for licensing reasons.                                   ##
##                                                          ##
## The code below loads our own readline as *the* readline  ##
## module and replaces any already pre-existing one in      ##
## sys.modules with our own. The reason for this to prevent ##
## any futher calls to "import readline" from importing the ##
## stock module.                                            ##
##                                                          ##
##############################################################
#
# Editline (libedit) seems to have a bug in it where it does
# not properly handle input/output streams that are not TTYs.
# As a result, it can completely screw up the terminal if the
# output is redirected.
# If both output streams are not TTYs, don't import the module
# and use non-readline methods to read the input stream.
use_readline = False
if sys.stdout.isatty() and sys.stderr.isatty():
    # The "xterm" terminal type come with meta-mode turned on and no way
    # to turn it off. The meta-mode causes the readline library to output
    # "ESC[?1034h" to stdout, which is un-desired.
    # Switch the terminal mode prior to loading the readline module
    if os.environ.get("TERM", "").find("xterm") != -1:
        os.environ["TERM"] = "vt100"
    try:
        import platform
        if (platform.dist()[0] == "redhat") and (platform.dist()[1] == "6.7"):
            import lib.readline.rhel67.readline as readline
        else:
            import lib.readline.default.readline as readline
    except ImportError:
        import lib.readline.libedit2.readline as readline
    sys.modules["readline"] = readline
    use_readline = True

# lib.readline needs to be imported before cmd because cmd
# imports readline internally and we want to prevent it
# pulling in the stock readline module
import cmd
from lib.hw import HwError, access2str
from lib.diags import *
from lib.logger import *

logger = initModuleLogging(__name__)

def getTerminalSize():
    data = (0, 0)
    for fd in (sys.stdin.fileno(), sys.stdout.fileno()):
        try:
            data = struct.unpack("hh", fcntl.ioctl(fd, termios.TIOCGWINSZ,
                                                   "abcd"))
        except: continue
        if data == (0, 0): continue
    if data == (0, 0): data = (25, 80)
    data = [int(os.environ.get("LINES", data[0])),
            int(os.environ.get("COLUMNS", data[1]))]
    return (data[1], data[0])

class UIConfig(object):
    types = {}
    def __setattr__(self, attr, value):
        if attr in self.types and type(value) != self.types[attr]:
            raise ValueError
        object.__setattr__(self, "prop_%s" % attr, value)
        if attr not in self.types: self.types[attr] = type(value)
    def __getattr__(self, attr):
        if attr in self.types:
            return self.types[attr](object.__getattribute__(self, "prop_%s"% attr))
        raise AttributeError
    def __getallprops__(self):
        return [x[5:] for x in dir(self) if x.startswith("prop_")]
    def get_props(self): return self.__getallprops__()


COMMAND = 0
COMPLETE = 1
HELP = 2

class CommandException(Exception): pass

class CommandLine(object, cmd.Cmd):
    init_done = False
    def __init__(self, completekey='tab', stdin=None, stdout=None, hw=None):
        logger.log(INFO, "hfidiags ui starting")
        cmd.Cmd.__init__(self, completekey, stdin, stdout)
        self.custom_commands = {}
        self.doc_leader = "All commands with unique first character have\n" + \
                          "an alias for that first character created automatically.\n"
        self.default_aliases = [("quit", "exit")]
        self.console = getTerminalSize()
        for sig in (signal.SIGWINCH, signal.SIGINT):
            signal.signal(sig, self.__signal_handler)
        self.aliases = {}
        self.config = UIConfig()
        self.set_default_config()
        self.interactive = True
        self.in_test_mode = False
        self.hw = hw
        self.__prompt = "diags"
        self.__methods__ = dir(self)
        self.prompt = ""
        if use_readline:
            delims = readline.get_completer_delims()
            for delim in os.sep + "[]-*~${}?+":
                delims = delims.replace(delim, "")
                readline.set_completer_delims(delims)
            # If the output stream is connected to a terminal, turn off printing
            # of control characters
            attrs = termios.tcgetattr(sys.stdout.fileno())
            attrs[3] &= ~termios.ECHOCTL
            termios.tcsetattr(sys.stdout.fileno(), termios.TCSANOW, attrs)
        else:
            self.use_rawinput = 0

        # Set this to True after the __init__ method has done it's work (see
        # comment below).
        self.init_done = True
        # Commands can only be registered after the init_done variable has been
        # set to True. Calling register_command() before that will result in
        # recursion within the __getattr__ method due to a broken hasattr()
        # in older versions of Python.
        self.register_command("exit", None, self.__exit_cmd, None, self.__exit_cmd_help)
        self.register_command("config", None, self.set_config, self.config_completer,
                              self.config_help)
        return

    def __getattr__(self, name):
        # Ignore requests for "aliases" until the __init__ calls has finished
        if name == "aliases" and not self_init_done:
            raise AttributeError("%s instance has not attribute '%s'" %
                                 (self.__class__.__name__, name))
        if self.aliases and isinstance(self.aliases, dict):
            try: cat, name = name.split("_", 1)
            except ValueError:
                raise AttributeError("%s instance has not attribute '%s'" %
                                     (self.__class__.__name__, name))
            fname = None
            # Check if we already have a defined alias
            if name in self.aliases:
                fname = '_'.join([cat, self.aliases[name]])
            else:
                # check for unique beginning
                methods = [x.split('_', 1)[1] for x in self.__methods__
                           if x.startswith(cat + "_")]
                matches = [x for x in methods if x.startswith(name)]
                if len(matches) == 1: fname = '_'.join([cat, matches[0]])
            if fname and hasattr(self, fname):
                func = getattr(self, fname)
                if callable(func): return func
        raise AttributeError("%s instance has not attribute '%s'" %
                             (self.__class__.__name__, name))

    def __setattr__(self, name, attr):
        if hasattr(self, "__methods__"): self.__methods__.append(name)
        super(CommandLine, self).__setattr__(name, attr)

    def __signal_handler(self, signo, frame):
        if signo == signal.SIGWINCH: self.console = getTerminalSize()
        elif signo == signal.SIGINT:
            # The processing of the keyboard interrupt is a little bit
            # special - the action depends on the call stack:
            #   1. if the program is in the command loop, the program just
            #      exits
            #   2. if it is executing a command, just the command is interrupted
            #      and control returns to the command loop
            #   3. if the command is in the middle of the pager, the signal is
            #      ignored so the underlying pager is quit properly
            # To accomplish this, we walk the call stack backwards looking for the
            # appropriate callers.
            while inspect.isframe(frame):
                if frame.f_code.co_name == "pager": break
                if frame.f_code.co_name in ("onecmd"): raise CommandException
                elif frame.f_code.co_name == "cmdloop":
                    # There is no clean way to return back to the loop in
                    # the cmdloop method
                    line = self.precmd("exit")
                    self.postcmd(self.onecmd(line), line)
                    if self.use_rawinput and self.completekey:
                        try:
                            # For some very odd reason, even though 'readline' is in
                            # the global scope, a direct reference to it fails. Until,
                            # understood, take it from sys.modules.
                            sys.modules["readline"].set_completer(self.old_completer)
                        except ImportError: pass
                    sys.exit(0)
                frame = frame.f_back

    def screen_width(self):
        return self.in_test_mode and 80 or self.console[0]

    def screen_height(self):
        return self.console[1]

    def __spacer(self, prefix, suffix, limit=-1):
        '''Insert proper amount of spaces between two strings.'''
        if limit == -1: limit = self.screen_width()
        if len(prefix) + len(suffix) + 1 > limit:
            return prefix+"\n"+" " * (limit - len(suffix)) + suffix
        return prefix + ' ' * (limit - len(prefix) - len(suffix)) + suffix

    # s - the string to be split
    # c - the line length at which to split the string
    # p - number of space to prefix the first line with
    # sp - number of spaces to prefix all other lines besides first
    # fignore - add prefix to first line
    def __line_split(self, s, c=80, p=0, sp=0, fignore=True):
        if not isinstance(s, basestring): return s
        if len(s) + p < c: return " " * p + s
        line, sep, first = "", c, fignore and True
        for rest in s.split("\n"):
            if not first: rest = " "*p + rest.rstrip()
            else: first = False
            if ' ' in rest:
                while rest:
                    if len(rest) <= sep:
                        line += rest
                        break
                    if rest[sep] != ' ':
                        while sep and rest[sep] != ' ': sep -= 1
                    line, rest, sep = \
                          line + (rest[:sep] and rest[:sep]+"\n" or ""), \
                          " " * (first and p or sp) + rest[sep+1:], c
            else:
                line += rest
            line += "\n"
        return line.rstrip()

    def __show_generic_error(self, prefix, text):
        self.__to_screen(self.__line_split(prefix + text, self.screen_width(), 3),
                         True)
        return

    def __dispatch_decorator(self, type, name, function):
        def __dispatch_command(line):
            if name in self.custom_commands and \
                   callable(self.custom_commands[name][0]):
                logger.log(INFO, "cmd: %s %s", name, line)
                cmdoutput = []
                ret = self.custom_commands[name][0](self, line, cmdoutput)
                if len(cmdoutput): self.show_cmd_output(cmdoutput)
                return ret
            return False
        def __dispatch_complete(text, line, si, ei):
            if name in self.custom_commands and \
                   callable(self.custom_commands[name][1]):
                return self.custom_commands[name][1](self, text, line[:ei], si, ei)
            return []
        def __dispatch_help():
            if name in self.custom_commands and \
                   callable(self.custom_commands[name][2]):
                help = self.custom_commands[name][2]()
                self.show_help_screen(name, help)
            return False
        if type == COMMAND: return __dispatch_command
        elif type == COMPLETE: return __dispatch_complete
        elif type == HELP: return __dispatch_help
        else: return None

    def __to_screen(self, text, error=False, paging=True):
        stream = (error and not self.interactive) and sys.stderr or sys.stdout
        if self.interactive and paging and self.config.output_paging and \
           len(text.split("\n")) > self.console[1]:
            pipepager(text, "less -E -F -K -R -X")
            return
        stream.write(text + "\n")
        # Have to flush the stream so regular output doesn't mix with
        # errors when executing a command file
        stream.flush()

    def __get_prompt(self):
        p = self.__prompt
        if self.hw and hasattr(self.hw, "unit"):
            p += "[hfi1_%s]" % (self.hw.unit == -1 and "-" or self.hw.unit)
        else: p += "[hfi1_-]"
        if self.prompt: p += " (" + self.prompt + ")"
        return p + " >> "

    def text2columns(self, words, as_columns=False):
        def to_slices(array, count=1):
            size = len(array) / count + (len(array) % count and 1 or 0)
            for x in range(0, len(array), size): yield array[x:x+size]
        if not list: return []
        width = self.screen_width()
        # throw away non-strings
        words = filter(lambda x: isinstance(x, str), words)
        # sort the words alphabetically
        words = sorted(words)
        space, num_columns = 1, 0
        while space > 0:
            space = width
            slices = list(to_slices(words, num_columns + 1))
            slice_len = 0
            for s in slices:
                s = sorted(s, lambda x, y: cmp(len(x), len(y)), reverse=True)
                slice_len += len(s[0])
            space -= slice_len + len(slices)
            if space < 0: break
            num_columns += 1

        columns = list(to_slices(words, num_columns))
        if as_columns == True: return columns
        formatted = []
        format = "".join(["%%-%ds" % (width / num_columns)] * num_columns)
        for idx in range(len(columns[0])):
            values = []
            for c in range(num_columns):
                if len(columns[c]) > idx: values.append(columns[c][idx])
                else: values.append("")
            formatted.append(format % tuple(values))
        return formatted

    def set_prompt(self, prompt=""):
        self.prompt = str(prompt)

    def enable_tests(self):
        self.in_test_mode = True

    def onecmd(self, line):
        line = line.strip()
        if line and line[0] == "#": return
        if line and line[-1] == ';': line = line[:-1]
        return super(CommandLine, self).onecmd(line)

    def set_default_config(self):
        # Repeat last command on "Enter"
        self.config.last_on_enter = True
        # Show decode bitfields
        self.config.show_fields = True
        # Show CSRs with 0 value
        self.config.show_zero_values = True
        # Show bitfields without any set bits
        self.config.show_zero_bitfield_values = True
        # Show bitfields as 64bit values or their respective width
        self.config.show_64bit_fields = False
        # Show bitfields when reading register(s)
        self.config.show_register_bitfields = True
        # Use the output pager or not
        self.config.output_paging = True

    def get_names(self):
        # Override the default method so it picks up custom registered
        # commands
        return [x for x in dir(self) if not x.startswith('_')]

    def emptyline(self):
        if self.config.last_on_enter and self.lastcmd:
            return self.onecmd(self.lastcmd)

    def default(self, line):
        self.__to_screen('*** Unknown syntax: %s'%line, True)

    # Override the super-class' cmdloop method to properly handle
    # KeyboardInterrupt exception
    def cmdloop(self, intro=None):
        self.preloop()
        if self.use_rawinput and self.completekey:
            try:
                # For some very odd reason, even though 'readline' is in
                # the global scope, a direct reference to it fails. Until,
                # understood, take it from sys.modules.
                self.old_completer = sys.modules["readline"].get_completer()
                sys.modules["readline"].set_completer(self.complete)
            except ImportError: pass
        try:
            if intro is not None: self.intro = intro
            if self.intro: self.stdout.write(str(self.intro)+"\n")
            stop = None
            while not stop:
                if self.cmdqueue: line = self.cmdqueue.pop(0)
                else:
                    prompt = self.__get_prompt()
                    if self.use_rawinput:
                        try: line = raw_input(prompt)
                        except EOFError: line = 'EOF'
                    else:
                        self.stdout.write(prompt)
                        self.stdout.flush()
                        line = self.stdin.readline()
                        if not len(line): line = 'EOF'
                        else: line = line.rstrip('\r\n')
                line = self.precmd(line)
                try: stop = self.onecmd(line)
                except CommandException: self.__to_screen("Command interrupted", True)
                stop = self.postcmd(stop, line)
            self.postloop()
        finally:
            if self.use_rawinput and self.completekey:
                try:
                    # See comment at top of function
                    sys.modules["readline"].set_completer(self.old_completer)
                except ImportError: pass

    def set_hw_object(self, obj):
        self.hw = obj
        method = getattr(self.hw, "set_print_func", None)
        if callable(method): method(self.ui_print)
        return True

    def format_numeric_value(self, value, width=64, bitfield=False):
        if isinstance(value, (int, long)):
            if not bitfield:
                if not self.config.show_zero_values and value == 0: return ""
                return ("0x%%0%02dx" % int(width / 4)) % value
            else:
                if not self.config.show_zero_bitfield_values and value == 0:
                    return ""
                if self.config.show_64bit_fields: return "0x%016x" % value
                else: return ("0x%%0%dx" % (int(width / 4) + (width % 4 > 0))) % \
				value
        return str(value)

    def format_register(self, reg, idx, elem, addr, value, fields=[]):
        lines, line = [], reg
        hr = self.hw.get_register_by_name(reg)
        if idx != -1 and hr and hr.count > 1: line += "[%d]" % idx
        if elem != -1 and hr and hr.array_size > 1: line += "[%d]" % elem
        valuestr = self.format_numeric_value(value, hr.width)
        if not valuestr: return ""
        addrstr = "%#-18x" % addr
        line += ' @ ' + addrstr
        line = self.__spacer(line, valuestr)
        lines.append(line)
        if fields and self.config.show_fields:
            lines.append('-' * self.screen_width())
            fsize = 0
            for field in fields:
                if len(field[0]) > fsize: fsize = len(field[0])
            fmt = "%%s.%%-%ds" % fsize
            for field in fields:
                fname, bits, shift, val = field
                valuestr = self.format_numeric_value(val, bits, True)
                if not valuestr: continue
                line = fmt % (reg, fname) + " ("
                if bits == 1: line += "%d):" % shift
                else: line += "%d-%d):" % (shift + bits - 1, shift)
                line = self.__spacer(line, valuestr)
                lines.append(line)
        return "\n".join(lines)

    def format_register_def(self, reg, addr, idx=0, fields=[]):
        lines, line = [], reg
        if idx: line += "[%d]" % idx
        addrstr = "@ %#010x" % addr
        line = self.__spacer(line, addrstr)
        lines.append(line)
        if fields:
            lines.append('-' * self.screen_width())
            line, i = "FIELD", "%9s %8s %18s %18s" % ("ACCESS", "SHIFT", "MASK", "SMASK")
            line = self.__spacer(line, i)
            lines.append(line)
            for field in fields:
                fname, access, shift, mask, smask = field
                line, i = fname, "%9s %8d %#18x %#18x" % \
                          (access2str(access), shift, mask, smask)
                line = self.__spacer(line, i)
                lines.append(line)
        return "\n".join(lines)

    def show_cmd_output(self, output_lines=[], separate=True, paging=True):
        complete = []
        if separate: complete.append("")
        if not isinstance(output_lines, list): output_lines = [output_lines]
        for line in output_lines:
            if line == SPACER: complete.append("-" * self.screen_width())
            elif line == HEADER: complete.append("=" * self.screen_width())
            elif line == EMPTY_LINE: complete.append("\n")
            elif isinstance(line, tuple):
                complete.append(self.__spacer(line[0], str(line[1])))
            else:
                if line is "": complete.append(line)
                else:
                    # __line_split strips the end newline. If the line ends with a
                    # new line, add it back
                    complete.append(self.__line_split(line,
                                                      self.screen_width()) + \
                                    (line[-1] == "\n" and "\n" or ""))
        if separate: complete.append("")
        complete = "\n".join(complete)
        # Don't allow double newlines at the end
        if complete.endswith("\n\n"): complete = complete[:-1]
        self.__to_screen(complete, False, paging)
        return

    def show_help_screen(self, cmd, help, tablen=8, sections=[]):
        doc_string = "\n\nFor more information, please refer to the documentation" + \
                     " found in /usr/share/hfidiags/docs"
        if sections: section_names = sections
        else: section_names = ["synopsis", "description", "arguments",
                               "returns", "known issues", "see also"]
        complete = ""
        for section in section_names:
            if section in help:
                if section == "arguments" and \
                   isinstance(help[section], list):
                    complete += "\n" + section.upper() + "\n"
                    for line in help[section]:
                        if isinstance(line, str):
                            complete += self.__line_split(line, self.screen_width(),
                                                          tablen, tablen, False) + "\n"
                        elif isinstance(line, list):
                            optlen = max([len(x[0]) for x in line])
                            fmt = "%%-%ds" % optlen
                            for opt in line:
                                complete += self.__line_split(fmt % opt[0] + " - " +
                                                              opt[1],
                                                              self.screen_width(),
                                                              tablen,
                                                              tablen + optlen + 3,
                                                              False) + "\n"
                elif section == "see also":
                    complete += "\n" + section.upper() + "\n" + \
                                self.__line_split(help[section] + doc_string,
                                                  self.screen_width(), tablen,
                                                  tablen, False) + "\n"
                else:
                    complete += "\n" + section.upper() + "\n" + \
                                self.__line_split(help[section], self.screen_width(),
                                                  tablen, tablen, False) + "\n"
        if cmd in self.aliases.values():
            alias = [k for k, v in self.aliases.items() if v == cmd][0]
            complete += "\nALIAS\n" + self.__line_split(alias, self.screen_width(),
                                                        tablen, tablen, False) + "\n"
        self.__to_screen(complete)
        return

    def ui_print(self, fmt, *args):
        self.show_cmd_output([fmt % args ], False)

    def error(self, text):
        self.__show_generic_error("ERROR: ", text)

    def hw_error(self, text):
        self.__show_generic_error("ERROR(HW): ", str(text))

    def __exit_cmd_help(self):
        return {"synopsis": "exit",
                "description":
                "Exit the application. Pressing Ctrl-C will also exit the application."}

    def __exit_cmd(self, obj, line, output):
        self.__to_screen("\nExiting application")
        for command in obj.custom_commands.keys():
            self.unregister_command(command)
        self.hw.close()
        return True

    def load_commands(self, path):
        module_path = os.sep.join([path, "commands"])
        if os.path.exists(module_path) and \
           os.path.isdir(module_path):
            for filename in os.listdir(module_path):
                if os.path.isfile(os.sep.join([module_path, filename])):
                    if not filename[-3:] == ".py": continue
                    modname = filename[:-3]
                    try:
                        fd, path, desc = imp.find_module(modname, [module_path])
                        module = imp.load_module(modname, fd, path, desc)
                        if hasattr(module, "__commands__"):
                            for com in module.__commands__:
                                if len(com) == 4:
                                    self.register_command(com[0], None, *com[1:])
                                elif len(com) == 5:
                                    self.register_command(*com)
                    except (ImportError, NameError): continue
                    finally:
                        if fd: fd.close()

    def register_command(self, name, alias, handler, completer=None, help=None):
        if not handler or not callable(handler): return False
        if not hasattr(self, "do_%s" % name):
            callback = self.__dispatch_decorator(COMMAND, name, handler)
            if callback: setattr(self, "do_%s" % name, callback)
            else: return False
            if completer:
                callback = self.__dispatch_decorator(COMPLETE, name, completer)
                if callback: setattr(self, "complete_%s" % name, callback)
            if help:
                callback = self.__dispatch_decorator(HELP, name, help)
                if callback: setattr(self, "help_%s" % name, callback)
            self.custom_commands[name] = (handler, completer, help)
            if alias and alias not in self.aliases: self.aliases[alias] = name
        return True

    def unregister_command(self, name):
        if name in self.custom_commands:
            for pre in ["help", "complete", "do"]:
                if hasattr(self, "%s_%s" % (pre, name)):
                    delattr(self, "%s_%s" % (pre, name))
            del self.custom_commands[name]
            return True
        return False

    def help_addressing(self):
        s = ["REGISTER SPECIFICATION AND ADDRESSING",
             "INDEX FORMATTING", "SCALAR REGISTER SPECIFICATION",
             "CONTEXT-INDEXED REGISTER SPECIFICATION",
             "CONTEXT- AND ARRAY-INDEX REGISTER SPECIFICATION", "BITFIELDS",
             "STMBOLIC REGISTER INDEXES", "GLOBALS/WILDCARDS"]
        a = {"REGISTER SPECIFICATION AND ADDRESSING":
             "All tool commands which accept register names as operands " + \
             "accept and use the following register specification. The " + \
             "information below is also provided as help text within the " + \
             "tool.\n\nThe register specification format can be used to " + \
             "reference Hardware registers by their symbolic names instead " + \
             "of address offsets. The register names are defined to match " + \
             "the Hardware Specification documents.\n\nThere are three types " + \
             "of Hardware registers - scalar, context-indexed, and context- " + \
             "and array-indexed. Scalar registers are registers which have " + \
             "only one instance defined in the Hardware. Context-index " + \
             "registers are registers which have one instance per Send " + \
             "and/or Receive context (depending on the section where they " + \
             "are defined). Context- and array-index registers are " + \
             "registers which have an array of instances for every Send " + \
             "and/or Receive context (depending on the section where the " + \
             "register is defined).\n\nThe Register Specification format " + \
             "defines a common way to reference any of the registers, " + \
             "independently of the type of register.",
             "INDEX FORMATTING":
             "Registers which support indexing (Context-indexed and " + \
             "context- and array-index) can be a single index or an index " + \
             "range. A single index will reference only the register or " + \
             "array instance indexed. A range will reference all registers " + \
             "or array instances between the starting and ending index, " + \
             "inclusively.\n\nBoth context indexes and array indexes " + \
             "support the above described format.",
             "SCALAR REGISTER SPECIFICATION":
             "Scalar registers are registers which have only one instance " + \
             "defined in the Hardware. These registers do not support any " + \
             "type of indexing and if indexing is used, an error will be " + \
             "reported.\n\nTo reference a scalar register, only the name " + \
             "of that register is required.\n\nExamples:\n\tread CceRevision",
             "CONTEXT-INDEXED REGISTER SPECIFICATION":
             "Context-index register are registers which have one instance " + \
             "per Send or Receive context.\n\nArray-only-index registers " + \
             "are registers which have multiple instances defined, however " + \
             "these instances are global for the entire chip. Examples of " + \
             "such registers are the counter registers.\n\nTo reference the " + \
             "register index for a particular context or a particular array " + \
             "element, a standard one-dimensional array indexing syntax is " + \
             "used:\n\n\t<register name>[<index>]\n\nThe index is optional. " + \
             "If the index is omitted, the command executed will perform " + \
             "its action on every register in the register set. For " + \
             "example: 'read RcvCtxtCtrl' will read and display the " + \
             "value of the RcvCtxtCtrl register for each of the Receive " + \
             "contexts defined by the Hardware.\n\nExamples:\n\tread " + \
             "SendCtxtCtrl[3]\n\tread RcvCtxtCtrl[1-4]",
             "CONTEXT- AND ARRAY-INDEX REGISTER SPECIFICATION":
             "Context- and array-index registers are registers which have " + \
             "a number of instances defined per Send or Receive context. " + \
             "If the Hardware defines N contexts of a certain type (Send " + \
             "or Receive) and M instances per context, then the register " + \
             "will have a total of NxM instances.\n\nTo reference such " + \
             "registers a standard two-dimensional array indexing syntax " + \
             "is used:\n\n\t<register name>[<index>][<index>]\n\nBoth the " + \
             "context index and the array index are optional. However, if " + \
             "only one index is present, it is assumed that it is the " + \
             "Context index. In other words, there is no support for " + \
             "short-circuiting the Context index of the register.\n\nAs " + \
             "with the Context-indexed register specification, if any " + \
             "of the indexes are missing, the operation will be carried " + \
             "out on the entire range.\n\nExamples:\n\tread " + \
             "RcvTidFlowTable[3][0]\n\tread RcvTidFlowTable[4-6][3]\n\t" + \
             "read RcvTidFlowTable[4-6][0-3]",
             "BITFIELDS":
             "All three types of register specifications support the use " + \
             "of bit fields. When bit fields are used, the commands are " + \
             "limited to performing their action only on the specified bit " + \
             "field. The rest of the register will be ignored.\n\nCurrently, " + \
             "only one bit field at a time can be specified. If multiple bit " + \
             "fields within a register need to be operated on, multiple " + \
             "commands need to be issued - one for each bit field.\n\nIn " + \
             "order to specify a register bit field, the following syntax is " + \
             "used:\n\n\t<register name>[<index>][<index>]\n\nwhere " + \
             "\"<register spec>\" is any of the three formats described " + \
             "above. Below are examples of this format:\n\n\tScalar registers: " + \
             "<register name>.<bit field>\n\tContext-indexed or " + \
             "Array-only-index: <register name>[<index>].<bit field>\n\t" + \
             "Context- and Array-indexed registers: <register " + \
             "name>[<index>][<index>].<bit field>",
             "SYMBOLIC REGISTER INDEXES":
             "Some Array-Only-Index CSRs defined in the HW specification " + \
             "include symbolic references to their indexes. Examples of " + \
             "such CSRs are the counter CSRs in the Send block. The tool " + \
             "supports use of these symbolic references. However, it does " + \
             "not do any special processing on the symbolic references such " + \
             "as verification and reference-CSR matching (whether the " + \
             "symbolic reference applies to the CSR). Therefore, these " + \
             "symbolic references are treated just like names for numeric " + \
             "indexes. This means that they could be used to index any " + \
             "single- or multi-dimensional CSR, provided that the CSR has " + \
             "sufficient depth in the dimension in question.",
             "GLOBALS/WILDCARDS":
             "The user interface supports several regular-expression-style " + \
             "wildcards, which can be used to address multiple registers " + \
             "and/or bit fields.\n\nWildcard:\n\t* - Matches any (zero or " + \
             "more) alpha-numeric set of characters.\n\t+ - Matches one or " + \
             "more alpha-numeric characters.\n\t? - Matches exactly one " + \
             "alpha-numeric character\n\t{n} - Matches exactly n " + \
             "alpha-numeric characters\n\t{n,m} - Matches between n and m " + \
             "alpha-numeric characters\n\nWhen using the above wildcards, " + \
             "all other register addressing options are still supported. " + \
             "This means that register ranges (both context and array " + \
             "indexes) can still be used, as well as, bit fields. " + \
             "Furthermore, the above wildcards can also be used on the " + \
             "register bit field names.\n\nWildcard character can be used " + \
             "multiple times within the same scope (register name or bit " + \
             "field name). For example, this could be used to get all " + \
             "register values for specific class of registers.\n\nWhen the " + \
             "wildcard characters are used for both register names and " + \
             "register bit fields, only register which match both criteria " + \
             "will be returned. This means that from the set of registers " + \
             "which match the register name pattern, only those who contain " + \
             "fields which match the bit field pattern will be returns.\n\n" + \
             "While the above wildcards a similar or the same as the " + \
             "wildcards used in most regular expressions, other regular " + \
             "expression special characters are not supported. This is " + \
             "especially true for \".\" and \"[]\" due to their special meaning " + \
             "in register addressing. Another important consideration is " + \
             "that, while the actual wildcard characters are supported, " + \
             "non-greedy matching combinations (supported by some regular " + \
             "expression syntaxes) are not supported. Therefore, a matching " + \
             "character combination of (for example) \"+?\" will not produce " + \
             "the desired result."
             }
        self.show_help_screen("addressing", a, sections=s)

    def set_config(self, obj, line, output):
        def is_true(val):
            return is_bool(val) and (val.lower() in ("true", "yes", "1", "on"))
        def is_bool(val):
            return (val.lower() in ("false", "no", "off", "0", "true", "yes", "on", "1"))

        words = line.split()
        props = self.config.get_props()
        if not words:
            output += ["UI Config", SPACER]
            for prop in props:
                output.append((prop+':', getattr(self.config, prop)))
        elif len(words) == 1:
            try: prop = getattr(self.config, words[0])
            except AttributeError:
                self.error("Configuration does not contain property '%s'" % words[0])
                return False
            output += ["UI Config", SPACER, (words[0]+':', getattr(self.config, words[0]))]
        else:
            if words[0] in props:
                if is_bool(words[1]): val = is_true(words[1])
                else: val = words[1]
                try: setattr(self.config, words[0], val)
                except ValueError:
                    self.error("Invalid value for configuration option '%s'" % words[0])
            else: self.error("Configuration does not contain property '%s'" % words[0])
        return

    def config_completer(self, obj, text, line, start, end):
        if text: return find_matches(text, self.config.get_props())
        else: return self.config.get_props()

    def config_help(self):
        return {"synopsis": "config [<option> [<value>]]",
                "description":
                "Set or show UI configuration options."}
